#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2025-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

import os
import numpy as np
from mystic.samplers import SparsitySampler
from mystic.monitors import Monitor, LoggingMonitor
from mystic.solvers import PowellDirectionalSolver
from mystic.termination import NormalizedChangeOverGeneration as NCOG
from mystic.tools import listify as tolist
from mystic.cache.archive import file_archive, read as get_db
from mystic.models import rosen as cost
from mystic.bounds import Bounds
bounds = Bounds(0,5,n=4)


def test_reset(s):
    si0 = s.iters()
    se0 = s.evals()
    ai0 = s._sampler._all_iters
    ae0 = s._sampler._all_evals
    ls0 = len(s._sampler._stepmon)
    st0 = s._sampler.Terminated(all=True)
    #print(st0)
    assert not any(st0)
    assert not si0
    assert not se0
    assert not any(ai0)
    assert not any(ae0)
    assert not ls0

    # reset non-terminated solvers [None]
    s.sample(if_terminated=False, reset_all=True)
    st1 = s._sampler.Terminated(all=True)
    si1 = s.iters()
    se1 = s.evals()
    ai1 = s._sampler._all_iters
    ae1 = s._sampler._all_evals
    ls1 = len(s._sampler._stepmon)
    #print(st1)
    assert not any(st1)
    assert si1
    assert se1
    assert not any(ai1)
    assert sum(ae1) == N
    assert ls1 == 1

    # reset non-terminated solvers [None]
    s.sample(if_terminated=False, reset_all=True)
    st2 = s._sampler.Terminated(all=True)
    si2 = s.iters()
    se2 = s.evals()
    ai2 = s._sampler._all_iters
    ae2 = s._sampler._all_evals
    ls2 = len(s._sampler._stepmon)
    #print(st1)
    assert not any(st2)
    assert si2 > si1
    assert se2 > se1
    assert not any(ai2)
    assert sum(ae2) == N
    assert ls2 == 1

    # reset regardless
    s.sample(if_terminated=None, reset_all=True)
    st3 = s._sampler.Terminated(all=True)
    si3 = s.iters()
    se3 = s.evals()
    ai3 = s._sampler._all_iters
    ae3 = s._sampler._all_evals
    ls3 = len(s._sampler._stepmon)
    #print(st3)
    assert not any(st3)
    assert si3 > si2
    assert se3 > se2
    assert not any(ai3)
    assert sum(ae3) == N
    assert ls3 == 1

    # never reset
    s.sample(if_terminated=all, reset_all=None)
    st4 = s._sampler.Terminated(all=True)
    si4 = s.iters()
    se4 = s.evals()
    ai4 = s._sampler._all_iters
    ae4 = s._sampler._all_evals
    ls4 = len(s._sampler._stepmon)
    #print(st4)
    assert not any(st4)
    assert si4 > si3
    assert se4 > se3
    assert sum(ai4) > sum(ai3)
    assert sum(ae4) > sum(ae3)
    assert ls4 > ls3

    # never reset
    s.sample(if_terminated=all, reset_all=None)
    st5 = s._sampler.Terminated(all=True)
    si5 = s.iters()
    se5 = s.evals()
    ai5 = s._sampler._all_iters
    ae5 = s._sampler._all_evals
    ls5 = len(s._sampler._stepmon)
    #print(st5)
    assert not any(st5)
    assert si5 > si4
    assert se5 > se4
    assert sum(ai5) > sum(ai4)
    assert sum(ae5) > sum(ae4)
    assert ls5 > ls4

    # sample until any terminated, never reset
    while not any(s._sampler.Terminated(all=True)):
        s.sample(if_terminated=any, reset_all=None)

    st6 = s._sampler.Terminated(all=True)
    si6 = s.iters()
    se6 = s.evals()
    ai6 = s._sampler._all_iters
    ae6 = s._sampler._all_evals
    ls6 = len(s._sampler._stepmon)
    #print(st6)
    assert any(st6)
    assert si6 > si5
    assert se6 > se5
    assert sum(ai6) > sum(ai5)
    assert sum(ae6) > sum(ae5)
    assert ls6 > ls5

    # terminated, but don't reset
    s.sample(if_terminated=any, reset_all=None)
    st7 = s._sampler.Terminated(all=True)
    si7 = s.iters()
    se7 = s.evals()
    ai7 = s._sampler._all_iters
    ae7 = s._sampler._all_evals
    ls7 = len(s._sampler._stepmon)
    #print(st7)
    assert any(st7)
    assert si7 > si6
    assert se7 > se6
    assert sum(ai7) == sum(ai6) + (N-sum(st6))
    assert sum(ae7) == sum(ae6) if all(st6) else sum(ae7) > sum(ae6)
    assert ls7 >= ls6 #XXX: better

    # terminated, but don't reset
    s.sample(if_terminated=all, reset_all=None)
    st8 = s._sampler.Terminated(all=True)
    si8 = s.iters()
    se8 = s.evals()
    ai8 = s._sampler._all_iters
    ae8 = s._sampler._all_evals
    ls8 = len(s._sampler._stepmon)
    #print(st8)
    assert any(st8)
    assert si8 > si7
    assert se8 > se7
    assert sum(ai8) == sum(ai7) + (N-sum(st7))
    assert sum(ae8) == sum(ae7) if all(st7) else sum(ae8) > sum(ae7)
    assert ls8 >= ls7 #XXX: better

    # terminated, and reset all
    s.sample(if_terminated=all, reset_all=True)
    st9 = s._sampler.Terminated(all=True)
    si9 = s.iters()
    se9 = s.evals()
    ai9 = s._sampler._all_iters
    ae9 = s._sampler._all_evals
    ls9 = len(s._sampler._stepmon)
    #print(st9)
    assert (not any(st9)) if all(st8) else any(st9)
    assert si9 > si8
    assert se9 > se8
    assert not any(ai9) if all(st8) else sum(ai9) > sum(ai8)
    assert sum(ae9) == N if all(st8) else sum(ae9) > sum(ae8)
    assert ls9 == 1 if all(st8) else ls9 >= ls8

    # sample until any terminated, reset all
    while not any(s._sampler.Terminated(all=True)):
        s.sample(if_terminated=any, reset_all=True)

    st10 = s._sampler.Terminated(all=True)
    si10 = s.iters()
    se10 = s.evals()
    ai10 = s._sampler._all_iters
    ae10 = s._sampler._all_evals
    ls10 = len(s._sampler._stepmon)
    #print(st10)
    assert any(st10)
    assert si10 >= si9 #XXX: better
    assert se10 >= se9 #XXX: better
    assert all(ai10)
    assert sum(ae10) > sum(ae9) if sum(st10) > sum(st9) else sum(ae10) == sum(ae9)
    assert ls10 > ls9 if sum(st10) > sum(st9) else ls10 == ls9

    # reset all regardless
    s.sample(if_terminated=None, reset_all=True)
    st11 = s._sampler.Terminated(all=True)
    si11 = s.iters()
    se11 = s.evals()
    ai11 = s._sampler._all_iters
    ae11 = s._sampler._all_evals
    ls11 = len(s._sampler._stepmon)
    #print(st11)
    assert not any(st11)
    assert si11 > si10
    assert se11 > se10
    assert not all(ai11)
    assert sum(ae11) == N
    assert ls11 == 1

    # reset all regardless
    s.sample(if_terminated=None, reset_all=True)
    st12 = s._sampler.Terminated(all=True)
    si12 = s.iters()
    se12 = s.evals()
    ai12 = s._sampler._all_iters
    ae12 = s._sampler._all_evals
    ls12 = len(s._sampler._stepmon)
    #print(st12)
    assert not any(st12)
    assert si12 > si11
    assert se12 > se11
    assert not all(ai12)
    assert sum(ae12) == N
    assert ls12 == 1


def test_init(s):
    s._reset_sampler()
    s.sample_until(terminated=any)
    t0 = s._sampler.Terminated(all=True)
    assert np.any(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    s.sample(None, True) # reset all solvers regardless
    tN = s._sampler.Terminated(all=True)
    assert not np.any(tN)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 != xN, regardless
    assert np.all([j if i else j for i,j in zip(tN, np.any(np.array(x0) - xN, axis=1))]) 
    s._reset_sampler()
    s.sample_until(terminated=any)
    t0 = s._sampler.Terminated(all=True)
    assert np.any(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    s.sample(None, None) # never reset regardless [NOTE: does nothing]
    tN = s._sampler.Terminated(all=True)
    assert np.any(tN)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 == xN, regardless
    assert np.all([~j if i else ~j for i,j in zip(tN, np.any(np.array(x0) - xN, axis=1))]) 
    t0 = s._sampler.Terminated(all=True)
    assert np.any(t0)
    s.sample(None, False) # reset terminated solvers regardless
    tN = s._sampler.Terminated(all=True)
    assert not np.any(tN)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    ## x0 != xN, if terminated
    assert np.all([j if i else ~j for i,j in zip(t0, np.any(np.array(x0) - xN, axis=1))])
    s.sample(False, True) # if none terminated, reset all solvers
    t0 = s._sampler.Terminated(all=True)
    assert not np.any(t0)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 != xN, if none terminated
    assert np.all([j if i else j for i,j in zip(tN, np.any(np.array(x0) - xN, axis=1))])
    s.sample(False, None) # if none terminated, never reset [NOTE: does nothing]
    tN = s._sampler.Terminated(all=True)
    assert not np.any(tN)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    ## x0 == xN, regardless
    assert np.all([~j if i else ~j for i,j in zip(t0, np.any(np.array(x0) - xN, axis=1))])
    s.sample(False, False) # if none terminated, reset terminated solvers (irrelevant)
    t0 = s._sampler.Terminated(all=True)
    assert not np.any(t0)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 == xN, regardless
    assert np.all([~j if i else ~j for i,j in zip(tN, np.any(np.array(x0) - xN, axis=1))])
    s._reset_sampler()
    s.sample_until(terminated=any)
    t0 = s._sampler.Terminated(all=True)
    assert np.any(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    s.sample(any, None) # if any terminated, never reset [NOTE: does nothing]
    tN = s._sampler.Terminated(all=True)
    assert np.any(tN)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 == xN, regardless
    assert np.all([~j if i else ~j for i,j in zip(tN, np.any(np.array(x0) - xN, axis=1))])
    s.sample(any, False) # if any terminated, reset terminated solvers
    t0 = s._sampler.Terminated(all=True)
    assert not np.any(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    ## x0 != xN, if any terminated
    assert np.all([j if i else ~j for i,j in zip(tN, np.any(np.array(x0) - xN, axis=1))])
    s._reset_sampler()
    s.sample_until(terminated=any)
    t0 = s._sampler.Terminated(all=True)
    assert np.any(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    s.sample(any, True) # if any terminated, reset all solvers
    tN = s._sampler.Terminated(all=True)
    assert not np.any(tN)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 != xN, if any terminated
    assert np.all([j if i else j for i,j in zip(t0, np.any(np.array(x0) - xN, axis=1))])
    s._reset_sampler()
    s.sample_until(terminated=all)
    t0 = s._sampler.Terminated(all=True)
    assert np.all(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    s.sample(all, None) # if all terminated, never reset [NOTE: does nothing]
    tN = s._sampler.Terminated(all=True)
    assert np.all(tN) 
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 == xN, regardless
    assert np.all([~j if i else ~j for i,j in zip(t0, np.any(np.array(x0) - xN, axis=1))])
    s.sample(all, True) # if all terminated, reset all solvers
    t0 = s._sampler.Terminated(all=True)
    assert not np.any(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    ## x0 != xN, if all terminated
    assert np.all([j if i else j for i,j in zip(tN, np.any(np.array(x0) - xN, axis=1))])
    s.sample(all, False) # if all terminated, reset terminated solvers
    tN = s._sampler.Terminated(all=True)
    assert not np.any(tN)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 != xN, if all terminated
    assert np.all([~j if i else ~j for i,j in zip(t0, np.any(np.array(x0) - xN, axis=1))])
    s._reset_sampler()
    s.sample_until(terminated=all)
    t0 = s._sampler.Terminated(all=True)
    assert np.all(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    s.sample(True, None) # if best terminated, never reset [NOTE: does nothing]
    tN = s._sampler.Terminated(all=True)
    assert np.all(tN)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 == xN, regardless
    assert np.all([~j if i else ~j for i,j in zip(t0, np.any(np.array(x0) - xN, axis=1))])
    s.sample(True, True) # if best terminated, reset all solvers
    tN = s._sampler.Terminated(all=True)
    assert not np.any(tN)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 != xN, if best terminated
    assert np.all([j if i else j for i,j in zip(t0, np.any(np.array(x0) - xN, axis=1))])
    s._reset_sampler()
    s.sample_until(terminated=all)
    t0 = s._sampler.Terminated(all=True)
    assert np.all(t0)
    x0 = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert x0 == s._init_solution
    s.sample(True, False) # if best terminated, reset terminated solvers
    tN = s._sampler.Terminated(all=True)
    assert not np.any(tN)
    xN = [si._stepmon._x[0] for si in s._sampler._allSolvers]
    if hasattr(s, '_init_solution'):
        assert xN == s._init_solution
    ## x0 != xN, if best terminated
    assert np.all([j if i else j for i,j in zip(t0, np.any(np.array(x0) - xN, axis=1))])


if __name__ == '__main__':
    nx = 4; ny = None
    N = 4
    kwds = dict(npts=N, maxiter=100, maxfun=1000, id=0)
    s = SparsitySampler(bounds, cost, **kwds)
    test_reset(s)

    kwds = dict(npts=N, maxiter=10, maxfun=1000, id=0)
    s = SparsitySampler(bounds, cost, **kwds)
    test_reset(s)

    test_init(s)



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2009-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

#######################################################################
# scaling and mpi info; also optimizer configuration parameters
# hard-wired: use DE solver, don't use mpi, F-F' calculation
# (similar to concentration.in)
#######################################################################
scale = 1.0
#XXX: <mpi config goes here>

npop = 20
maxiter = 1000
maxfun = 1e+6
convergence_tol = 1e-4
crossover = 0.9
percent_change = 0.9


#######################################################################
# the model function
# (similar to Simulation.cpp)
#######################################################################
from surrogate import marc_surr as model


#######################################################################
# the subdiameter calculation
# (similar to driver.sh)
#######################################################################
def costFactory(i):
  """a cost factory for the cost function"""

  def cost(rv):
    """compute the diameter as a calculation of cost

  Input:
    - rv -- 1-d array of model parameters

  Output:
    - diameter -- scale * | F(x) - F(x')|**2
    """

    # prepare x and xprime
    params = rv[:-1]                         #XXX: assumes Xi' is at rv[-1]
    params_prime = rv[:i]+rv[-1:]+rv[i+1:-1] #XXX: assumes Xi' is at rv[-1]

    # get the F(x) response
    Fx = model(params)

    # get the F(x') response
    Fxp = model(params_prime)

    # compute diameter
    return -scale * (Fx - Fxp)**2

  return cost


#######################################################################
# the differential evolution optimizer
# (replaces the call to dakota)
#######################################################################
def optimize(cost,lb,ub):
  from mystic.solvers import DifferentialEvolutionSolver2
  from mystic.termination import CandidateRelativeTolerance as CRT
  from mystic.strategy import Best1Exp
  from mystic.monitors import VerboseMonitor, Monitor
  from mystic.tools import random_seed

  random_seed(123)

 #stepmon = VerboseMonitor(100)
  stepmon = Monitor()
  evalmon = Monitor()

  ndim = len(lb) # [(1 + RVend) - RVstart] + 1

  solver = DifferentialEvolutionSolver2(ndim,npop)
  solver.SetRandomInitialPoints(min=lb,max=ub)
  solver.SetStrictRanges(min=lb,max=ub)
  solver.SetEvaluationLimits(maxiter,maxfun)
  solver.SetEvaluationMonitor(evalmon)
  solver.SetGenerationMonitor(stepmon)

  tol = convergence_tol
  solver.Solve(cost,termination=CRT(tol,tol),strategy=Best1Exp, \
               CrossProbability=crossover,ScalingFactor=percent_change)

  print("solved: %s" % solver.bestSolution)
  diameter_squared = -solver.bestEnergy / scale  #XXX: scale != 0
  func_evals = solver.evaluations
  return diameter_squared, func_evals


#######################################################################
# loop over model parameters to calculate concentration of measure
# (similar to main.cc)
#######################################################################
def UQ(start,end,lower,upper):
  diameters = []
  function_evaluations = []
  total_func_evals = 0
  total_diameter = 0.0

  for i in range(start,end+1):
    lb = lower[start:end+1] + [lower[i]]
    ub = upper[start:end+1] + [upper[i]]
  
    #construct cost function and run optimizer
    cost = costFactory(i)
    subdiameter, func_evals = optimize(cost,lb,ub) #XXX: no initial conditions

    function_evaluations.append(func_evals)
    diameters.append(subdiameter)

    total_func_evals += function_evaluations[-1]
    total_diameter += diameters[-1]

  print("subdiameters (squared): %s" % diameters)
  print("diameter (squared): %s" % total_diameter)
  print("func_evals: %s => %s" % (function_evaluations, total_func_evals))

  return total_diameter


#######################################################################
# statistics
# probability mass, expectation, mean, diameter, McDiarmid
#######################################################################
from mystic.math.stats import volume, prob_mass, mean, mcdiarmid_bound
from mystic.math.integrate import integrate as expectation_value


#######################################################################
# rank, bounds, and restart information 
# (similar to concentration.variables)
#######################################################################
if __name__ == '__main__':
  from math import sqrt

  function_name = "marc_surr"
  lower_bounds = [60.0, 0.0, 2.1]
  upper_bounds = [105.0, 30.0, 2.8]
# h = thickness = [60,105]
# a = obliquity = [0,30]
# v = speed = [2.1,2.8]

  RVstart = 0; RVend = 2
  RVmax = len(lower_bounds)

  # when not a random variable, set the value to the lower bound
  for i in range(0,RVstart):
    upper_bounds[i] = lower_bounds[i]
  for i in range(RVend+1,RVmax):
    upper_bounds[i] = lower_bounds[i]

  lbounds = lower_bounds[RVstart:1+RVend]
  ubounds = upper_bounds[RVstart:1+RVend]

  print("...SETTINGS...")
  print("npop = %s" % npop)
  print("maxiter = %s" % maxiter)
  print("maxfun = %s" % maxfun)
  print("convergence_tol = %s" % convergence_tol)
  print("crossover = %s" % crossover)
  print("percent_change = %s" % percent_change)
  print("..............\n\n")

  print(" model: f(x) = %s(x)" % function_name)
  param_string = "["
  for i in range(RVmax): 
    param_string += "'x%s'" % str(i+1)
    if i == (RVmax - 1):
      param_string += "]"
    else:
      param_string += ", "

  print(" parameters: %s" % param_string)
  print(" lower bounds: %s" % lower_bounds)
  print(" upper bounds: %s" % upper_bounds)
# print(" ...")
 #cuboid_volume = volume(lower_bounds,upper_bounds)
  cuboid_volume = volume(lbounds,ubounds)
  probability_mass = prob_mass(cuboid_volume,cuboid_volume)
  print(" probability mass: %s" % probability_mass)
  expectation = expectation_value(model,lower_bounds,upper_bounds)
  print(" expectation: %s" % expectation)
  mean_value = mean(expectation,cuboid_volume)
  print(" mean value: %s" % mean_value)
  diameter = UQ(RVstart,RVend,lower_bounds,upper_bounds)
  mcdiarmid = mcdiarmid_bound(mean_value,sqrt(diameter))
  print("McDiarmid bound: %s" % mcdiarmid)

 #if RVstart != 0 or RVend != 2: # abort when not a 3-D problem
 #  break #FIXME: break? or exit? or pass/else? or ???

  weighted_bound = []
  sanity = []
  lb = []
  ub = []
# subdivisions at h=100, a=20, v=2.2
  lb.append([60.0, 0.0, 2.1])
  ub.append([100.0, 20.0, 2.2])

  lb.append([60.0, 0.0, 2.2])
  ub.append([100.0, 20.0, 2.8])

  lb.append([60.0, 20.0, 2.1])
  ub.append([100.0, 30.0, 2.2])

  lb.append([60.0, 20.0, 2.2])
  ub.append([100.0, 30.0, 2.8])

  lb.append([100.0, 0.0, 2.1])
  ub.append([105.0, 20.0, 2.2])

  lb.append([100.0, 0.0, 2.2])
  ub.append([105.0, 20.0, 2.8])

  lb.append([100.0, 20.0, 2.1])
  ub.append([105.0, 30.0, 2.2])

  lb.append([100.0, 20.0, 2.2])
  ub.append([105.0, 30.0, 2.8])

  for i in range(len(lb)):
    print("\n")
    print(" lower bounds: %s" % lb[i])
    print(" upper bounds: %s" % ub[i])
#   print(" ...")
    subcuboid_volume = volume(lb[i],ub[i])
    sub_prob_mass = prob_mass(subcuboid_volume,cuboid_volume)
    sanity.append(sub_prob_mass)
    print(" probability mass: %s" % sub_prob_mass)
    expect_value = expectation_value(model,lb[i],ub[i])
    print(" expectation: %s" % expect_value)
    sub_mean_value = mean(expect_value,subcuboid_volume)
    print(" mean value: %s" % sub_mean_value)
    sub_diameter = UQ(RVstart,RVend,lb[i],ub[i])
    sub_mcdiarmid = mcdiarmid_bound(sub_mean_value,sqrt(sub_diameter))
    print("McDiarmid bound: %s" % sub_mcdiarmid)
    weighted_bound.append(sub_prob_mass * sub_mcdiarmid)

  # compare weighted to McDiarmid
  print("\n\n..............")
  p_mcdiarmid = probability_mass * mcdiarmid
  print("McDiarmid: %s" % p_mcdiarmid)
  weighted = sum(weighted_bound)
  print("weighted McDiarmid: %s" % weighted)
  try:
    print("relative change: %s" % (weighted / p_mcdiarmid))
  except ZeroDivisionError:
    pass
  assert sum(sanity) == probability_mass
  



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2018-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.solvers import DifferentialEvolutionSolver
from mystic.models import rosen
from mystic.tools import solver_bounds
from mystic.termination import ChangeOverGeneration as COG, Or, CollapseCost
from mystic.monitors import VerboseLoggingMonitor as Monitor

solver = DifferentialEvolutionSolver(3,40)
solver.SetRandomInitialPoints([-100]*3, [100]*3)

mon = Monitor(1)
options = dict(limit=1.95, samples=50, clip=False)
mask = {} #solver_bounds(solver)
stop = Or(CollapseCost(mask=mask,**options), COG(generations=50))

solver.SetGenerationMonitor(mon)
solver.SetTermination(stop)
solver.SetObjective(rosen)
solver.Solve()

print(solver.Terminated(info=True))
print('%s @' % solver.bestEnergy)
print(solver.bestSolution)



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
exercises the Metropolis-Hastings algorithm
"""

from mystic.metropolis import metropolis_hastings
from numpy import mean, cov

if __name__=='__main__':
    from numpy import random
    import time

    def prop(x):
        return random.normal(x, 1)

    def target(x):
        if x <0 or x > 3:
            return 0
        return 1. + 0.31831/(0.0025 + (x-1)**2) + 45. *x - 10. * x**2

    t1 = time.time()
    x = [1]
    for i in range(100000):
        x.append(metropolis_hastings(prop, target, x[-1]) )
    t2 = time.time()
    print('Metropolis took %0.3f ms' % ((t2-t1)*1000 ))

    import matplotlib.pyplot as plt
    plt.hist(x,20)
    plt.show()


# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
OUQ classes for calculating bounds on statistical quantities
"""
import mystic.cache
from mystic.cache import cached
from mystic.cache.archive import dict_archive, file_archive, read
from mystic.math import almostEqual
from mystic.math.discrete import product_measure
from mystic.math.samples import random_samples
from mystic.solvers import DifferentialEvolutionSolver2
from mystic.monitors import Monitor
from ouq_models import WrapModel


class BaseOUQ(object): #XXX: redo with a "Solver" interface, like ensemble?

    def __init__(self, model, bounds, **kwds):
        """OUQ model for a statistical quantity

    Input:
        model: function of the form y = model(x, axis=None)
        bounds: mystic.bounds.MeasureBounds instance

    Additional Input:
        samples: int, number of samples (used for non-deterministic models)
        penalty: function of the form y' = penalty(x)
        constraint: function of the form x' = constraint(x)
        xvalid: function returning True if x == x', given constraint
        cvalid: function similar to xvalid, but with product_measure input
        map: map instance, to evaluate model on the product_measure in parallel

    NOTE: when y is multivalued models must be a UQModel with ny != None
        """
        self.npts = bounds.n  # (2,1,1)
        self.lb = bounds.lower
        self.ub = bounds.upper
        #NOTE: for *_bound(axis=None) to work, requires ny != None
        #if not isinstance(model, WrapModel.mro()[-2]):
        #    model = WrapModel(model=model)
        self.model = model
        self.axes = getattr(model, 'ny', None) #FIXME: in kwds?, ny
        rnd = getattr(model, 'rnd', True) #FIXME: ?, rnd
        self.samples = kwds.get('samples', None) if rnd else None
        self.map = None if self.samples is None else kwds.get('map', None)
        self.penalty = kwds.get('penalty', lambda rv:0.)
        self.constraint = kwds.get('constraint', lambda rv:rv)
        self.xvalid = kwds.get('xvalid', lambda rv:True)
        self.cvalid = kwds.get('cvalid', lambda c:True)
        self._invalid = float('inf') #XXX: need to save state?
        self.kwds = {} #XXX: good idea???
        self._upper = {} # saved solver instances for upper bounds
        self._lower = {} # saved solver instances for lower bounds
        self._expect = {} # saved solver instances for expected value
        self._ave = {} # saved expected value from sampling
        self._var = {} # saved expected variance from sampling
        self._err = {} # saved misfit to sampled expected value
        self._cost = {} # saved cache of the most recent cost (for penalty)
        self._pts = {} # saved sampled {in:out} objective #XXX: out only? ''?
        return

    # --- extrema ---
    def expected(self, axis=None, **kwds):
        """find the expected value of the statistical quantity

    Input:
        axis: int, the index of y on which to find value (all, by default)
        axmap: map instance, to execute each axis in parallel (None, by default)
        instance: bool, if True, return the solver instance (False, by default)

    Additional Input:
        solver: mystic.solver instance [default: DifferentialEvolutionSolver2]
        npop: population size [default: None]
        id: a unique identifier for the solver [default: None]
        nested: mystic.solver instance [default: None], for ensemble solvers
        x0: initial parameter guess [default: use RandomInitialPoints]
        maxiter: max number of iterations [default: defined in solver]
        maxfun: max number of objective evaluations [default: defined in solver]
        evalmon: mystic.monitor instance [default: Monitor], for evaluations
        stepmon: mystic.monitor instance [default: Monitor], for iterations
        map: pathos map instance for solver.SetMapper [default: None]
        save: iteration frequency to save solver [default: None]
        opts: dict of configuration options for solver.Solve [default: {}]

    Further Input:
        archive: the archive (str name for a new archive, or archive instance)
        dist: a mystic.tools.Distribution instance (or list of Distributions)
        tight: if True, apply bounds concurrent with other constraints
        clip: if True, clip at bounds, else resample [default = False]
        smap: map instance, to sample the quantity in parallel [default = None]
        npts: maximum number of sample points [default = 10000]
        ipts: number of sample points per iteration [default = npts]
        iter: number of iterations to stop after if change < itol [default = 1]
        itol: stop if change in value < itol over iter [default = 1e-15]

    Returns:
        expected value of the statistical quantity, for the specified axis

    NOTE: by default, npts samplings of the expected value will be drawn,
        where npts is large. When the randomness of the model is expected to
        be small, the expected value can often be found quickly by setting
        ipts=1 and iter to a small number. Note however, that iterations are
        serial, while samplings within an iteration can utilize a parallel map.

    NOTE: an optimization is used to solve for inputs that yield the expected
        value, where the objective includes a penalty targeted to stop at the
        sampled mean. The additional penalty is 'linear_equality' with a
        strength, k, set equal to the sampled mean. The additional termination
        condition is 'VTR(ftol, target)', where ftol=1e-16 for models with
        no randomness, and ftol is the sampled variance for models with
        randomness. Target is the sampled mean. If only the expected value is
        of interest, setting `instance=None` will skip the optimization.
        """
        #NOTE: kwds(verbose, reducer) undocumented
        full = kwds.pop('instance', False)
        #self._expect.clear() #XXX: good idea?
        # use __call__ to get expected value [float or tuple(float)]
        from mystic.monitors import VerboseMonitor, Null
        verbose = kwds.pop('verbose', False) #NOTE: used for debug
        label = self.__class__.__name__
        mon = VerboseMonitor(1,None,label=label) if verbose else Null()
        db = kwds.pop('archive', None)
        if db is None: db = file_archive()
        elif type(db) in (str, (u''.__class__)):
            db = read(db, type=file_archive)
        # populate monitor with existing archive data
        m = Monitor()
        xs = list(db.values())
        N = size = ys = len(xs)
        if size:
            xtype = type(xs[-1]) # is tuple or float
            multi = hasattr(xtype, '__len__')
            import numpy as np
            xs = np.array(xs)
            var = xtype(xs.var(axis=0)) if multi else xs.var()
            if xs.ndim == 1: xs = np.atleast_2d(xs).T
            ys = np.arange(1, size+1)
            m._x = xs = (xs.cumsum(axis=0).T/ys).T.tolist()
            m._y = ys = ys.tolist()
            ave = xtype(xs[-1]) if multi else xs[-1][0]
            if verbose: print("%s: %s @ %s" % (label, ave, N))
        else:
            var = ave = float('nan') #XXX: good idea?
            multi = self.axes is not None
            if multi:
                var = ave = (ave,) * self.axes
        # sample expected value iteratively until termination met
        #reducer = kwds.pop('reducer', None) #XXX: throw error if provided?
        npts = kwds.pop('npts', 10000)
        ipts = kwds.pop('ipts', None)
        niter = kwds.pop('iter', None)
        itol = kwds.pop('itol', None)
        if npts is None: npts = float('inf')
        if ipts is None: ipts = npts
        if niter is None: niter = 1
        if itol is None: itol = 1e-15
        if ipts == float('inf') or ipts < 1:
            msg = 'ipts must be a positive integer'
            raise ValueError(msg)
        if npts < ipts:
            msg = 'ipts must be less than npts'
            raise ValueError(msg)
        smap = kwds.pop('smap', None)
        kwds_ = {} # kwds for _expected
        keys = ('reducer','dist','tight','clip') # keys for __call__
        kwds = {k:v for k,v in kwds.items() if k in keys or kwds_.update({k:v})}

        import numpy as np
        from mystic.abstract_solver import AbstractSolver
        from mystic.termination import CollapseAt
        solver = AbstractSolver(self.axes or 1) #XXX: axis?
        solver.SetTermination(CollapseAt(tolerance=itol, generations=niter))
        solver.SetEvaluationLimits(maxiter=npts-1) # max num of samples
        solver.SetGenerationMonitor(m)
        while not solver.Terminated() and N != npts:
            N += ipts                       #XXX: or self.axes?
            N = min(npts, N)
            ave = self.__call__(archive=db, axis=axis, npts=N, map=smap, **kwds)
            # update stepmon with the sampled values
            xs = list(db.values())
            xtype = type(xs[-1]) # is tuple or float
            multi = hasattr(xtype, '__len__')
            xs = np.array(xs)
            var = xtype(np.nanvar(xs, axis=0)) if multi else np.nanvar(xs)
            if xs.ndim == 1: xs = np.atleast_2d(xs).T
            ys = np.arange(1, xs.shape[0]+1)
            y_ = ys - np.cumsum(np.isnan(xs), axis=0).T # nan-adjusted
            solver._stepmon._x = xs = (np.nancumsum(xs, axis=0).T/y_).T.tolist()
            solver._stepmon._y = ys = ys.tolist()
            del y_; ave = xtype(xs[-1]) if multi else xs[-1][0]
            mon(N, ave) #XXX: or use (ave, N)?

        if verbose and len(m) > size:
            print(solver.Terminated(info=True)) #XXX: also disp=True?
            print("%s: %s @ %s" % (label, ave, N))

        # downselect ave to the specified axis, where relevant
        if isinstance(ave, tuple): # self.axes is not None
            for i,me in enumerate(ave):
                self._ave[i] = me
                self._var[i] = var[i]
                self._expect[i] = None
                self._err[i] = None
        else:
            ax = None if self.axes is None else axis
            self._ave[ax] = ave
            self._var[ax] = var
            self._expect[ax] = None
            self._err[ax] = None

        if full is None: # short-circuit the solver
            return ave
        # solve for params that yield expected value
        if self.axes is None or axis is not None:
            # solve for expected value of objective (in measure space)
            solver = self._expected(axis, **kwds_)
            ax = None if self.axes is None else axis
            self._expect[ax] = solver
            self._err[ax] = me = abs(self._ave[ax] - solver.bestEnergy)
            if verbose:
                print("%s: misfit = %s, var = %s" % (ax, me, self._var[ax]))
            if full: return solver
            return ave #NOTE: within misfit of solver.bestEnergy
        # else axis is None
        solvers = self._expected(axis, **kwds_)
        for ax,solver in enumerate(solvers):
            self._expect[ax] = solver
            self._err[ax] = me = abs(self._ave[ax] - solver.bestEnergy)
            if verbose:
                print("%s: misfit = %s, var = %s" % (ax, me, self._var[ax]))
        if full: return solvers
        return ave #NOTE: within misfit of solver.bestEnergy

    def _expected(self, axis=None, **kwds):
        """find the expected value of the statistical quantity

    Input:
        axis: int, the index of y on which to find value (all, by default)
        axmap: map instance, to execute each axis in parallel (None, by default)

    Additional Input:
        solver: mystic.solver instance [default: DifferentialEvolutionSolver2]
        npop: population size [default: None]
        id: a unique identifier for the solver [default: None]
        nested: mystic.solver instance [default: None], for ensemble solvers
        x0: initial parameter guess [default: use RandomInitialPoints]
        maxiter: max number of iterations [default: defined in solver]
        maxfun: max number of objective evaluations [default: defined in solver]
        evalmon: mystic.monitor instance [default: Monitor], for evaluations
        stepmon: mystic.monitor instance [default: Monitor], for iterations
        map: pathos map instance for solver.SetMapper [default: None]
        save: iteration frequency to save solver [default: None]
        opts: dict of configuration options for solver.Solve [default: {}]

    Returns:
        solver instance with solved expected value of the statistical quantity
        """
        axmap = kwds.pop('axmap', map) #NOTE: was _ThreadPool.map w/ join
        if axmap is None: axmap = map
        self.kwds.update(**kwds) #FIXME: good idea???
        if self.axes is None or axis is not None:
            #FIXME: enable user-provided (kpen,ftol)?
            # set penalty at same scale as expected value
            kpen = self._ave[axis]
            from mystic.penalty import linear_equality
            penalty = linear_equality(lambda rv, ave: abs(ave - self._cost[axis]), kwds={'ave':self._ave[axis]}, k=kpen)(lambda rv: 0.)
            # stop at exact cost, however if noisy stop within variance
            ftol = 1e-16 if self.samples is None else self._var[axis]
            ftol = self.kwds.get('ftol', ftol) #FIXME: doc user provided ftol
            from mystic.termination import VTR
            stop = VTR(ftol, self._ave[axis])
            # solve for expected value of objective (in measure space)
            def objective(rv):
                cost = self._cost[axis] = self.objective(rv, axis)
                return cost
            #self._invalid, invalid = float('nan'), self._invalid
            solver = self.solve(objective, penalty=penalty, stop=stop)
            #self._invalid = invalid
            return solver
        # else axis is None
        expected = tuple(axmap(self._expected, range(self.axes)))
        return expected #FIXME: don't accept "uphill" moves?

    def upper_bound(self, axis=None, **kwds):
        """find the upper bound on the statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)
        axmap: map instance, to execute each axis in parallel (None, by default)
        instance: bool, if True, return the solver instance (False, by default)

    Additional Input:
        solver: mystic.solver instance [default: DifferentialEvolutionSolver2]
        npop: population size [default: None]
        id: a unique identifier for the solver [default: None]
        nested: mystic.solver instance [default: None], for ensemble solvers
        x0: initial parameter guess [default: use RandomInitialPoints]
        maxiter: max number of iterations [default: defined in solver]
        maxfun: max number of objective evaluations [default: defined in solver]
        evalmon: mystic.monitor instance [default: Monitor], for evaluations
        stepmon: mystic.monitor instance [default: Monitor], for iterations
        map: pathos map instance for solver.SetMapper [default: None]
        save: iteration frequency to save solver [default: None]
        opts: dict of configuration options for solver.Solve [default: {}]

    Returns:
        upper bound on the statistical quantity, for the specified axis
        """
        full = kwds.pop('instance', False)
        #self._upper.clear() #XXX: good idea?
        if self.axes is None or axis is not None:
            # solve for upper bound of objective (in measure space)
            solver = self._upper_bound(axis, **kwds)
            self._upper[None if self.axes is None else axis] = solver
            if full: return solver
            return solver.bestEnergy
        # else axis is None
        solvers = self._upper_bound(axis, **kwds)
        for i,solver in enumerate(solvers):
            self._upper[i] = solver
        if full: return solvers
        return tuple(i.bestEnergy for i in solvers)

    def _upper_bound(self, axis=None, **kwds):
        """find the upper bound on the statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)
        axmap: map instance, to execute each axis in parallel (None, by default)

    Additional Input:
        solver: mystic.solver instance [default: DifferentialEvolutionSolver2]
        npop: population size [default: None]
        id: a unique identifier for the solver [default: None]
        nested: mystic.solver instance [default: None], for ensemble solvers
        x0: initial parameter guess [default: use RandomInitialPoints]
        maxiter: max number of iterations [default: defined in solver]
        maxfun: max number of objective evaluations [default: defined in solver]
        evalmon: mystic.monitor instance [default: Monitor], for evaluations
        stepmon: mystic.monitor instance [default: Monitor], for iterations
        map: pathos map instance for solver.SetMapper [default: None]
        save: iteration frequency to save solver [default: None]
        opts: dict of configuration options for solver.Solve [default: {}]

    Returns:
        solver instance with solved upper bound on the statistical quantity
        """
        axmap = kwds.pop('axmap', map) #NOTE: was _ThreadPool.map w/ join
        if axmap is None: axmap = map
        self.kwds.update(**kwds) #FIXME: good idea???
        if self.axes is None or axis is not None:
            # solve for upper bound of objective (in measure space)
            self._invalid *= -1
            solver = self.solve(lambda rv: -self.objective(rv, axis))
            self._invalid *= -1
            return solver
        # else axis is None
        upper = tuple(axmap(self._upper_bound, range(self.axes)))
        return upper #FIXME: don't accept "uphill" moves?

    def lower_bound(self, axis=None, **kwds):
        """find the lower bound on the statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)
        axmap: map instance, to execute each axis in parallel (None, by default)
        instance: bool, if True, return the solver instance (False, by default)

    Additional Input:
        solver: mystic.solver instance [default: DifferentialEvolutionSolver2]
        npop: population size [default: None]
        id: a unique identifier for the solver [default: None]
        nested: mystic.solver instance [default: None], for ensemble solvers
        x0: initial parameter guess [default: use RandomInitialPoints]
        maxiter: max number of iterations [default: defined in solver]
        maxfun: max number of objective evaluations [default: defined in solver]
        evalmon: mystic.monitor instance [default: Monitor], for evaluations
        stepmon: mystic.monitor instance [default: Monitor], for iterations
        map: pathos map instance for solver.SetMapper [default: None]
        save: iteration frequency to save solver [default: None]
        opts: dict of configuration options for solver.Solve [default: {}]

    Returns:
        lower bound on the statistical quantity, for the specified axis
        """
        full = kwds.pop('instance', False)
        #self._lower.clear() #XXX: good idea?
        if self.axes is None or axis is not None:
            # solve for lower bound of objective (in measure space)
            solver = self._lower_bound(axis, **kwds)
            self._lower[None if self.axes is None else axis] = solver
            if full: return solver
            return solver.bestEnergy
        # else axis is None
        solvers = self._lower_bound(axis, **kwds)
        for i,solver in enumerate(solvers):
            self._lower[i] = solver
        if full: return solvers
        return tuple(i.bestEnergy for i in solvers)

    def _lower_bound(self, axis=None, **kwds):
        """find the lower bound on the statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)
        axmap: map instance, to execute each axis in parallel (None, by default)

    Additional Input:
        solver: mystic.solver instance [default: DifferentialEvolutionSolver2]
        npop: population size [default: None]
        id: a unique identifier for the solver [default: None]
        nested: mystic.solver instance [default: None], for ensemble solvers
        x0: initial parameter guess [default: use RandomInitialPoints]
        maxiter: max number of iterations [default: defined in solver]
        maxfun: max number of objective evaluations [default: defined in solver]
        evalmon: mystic.monitor instance [default: Monitor], for evaluations
        stepmon: mystic.monitor instance [default: Monitor], for iterations
        map: pathos map instance for solver.SetMapper [default: None]
        save: iteration frequency to save solver [default: None]
        opts: dict of configuration options for solver.Solve [default: {}]

    Returns:
        solver instance with solved lower bound on the statistical quantity
        """
        axmap = kwds.pop('axmap', map) #NOTE: was _ThreadPool.map w/ join
        if axmap is None: axmap = map
        self.kwds.update(**kwds) #FIXME: good idea???
        if self.axes is None or axis is not None:
            # solve for lower bound of objective (in measure space)
            return self.solve(lambda rv: self.objective(rv, axis))
        # else axis is None
        lower = tuple(axmap(self._lower_bound, range(self.axes)))
        return lower #FIXME: don't accept "uphill" moves?

    # --- func ---
    def objective(self, rv, axis=None):
        """calculate the statistical quantity, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the statistical quantity for the specified axis

    NOTE:
        respects constraints on input parameters and product measure
        """
        return NotImplemented

    def __call__(self, axis=None, reducer=None, **kwds):
        """apply the reducer to the sampled statistical quantity

    Input:
        axis: int, the index of y on which to find quantity (all, by default)
        reducer: function, reduces a list to a single value (mean, by default)

    Further Input:
        archive: the archive (str name for a new archive, or archive instance)
        dist: a mystic.tools.Distribution instance (or list of Distributions)
        tight: if True, apply bounds concurrent with other constraints
        clip: if True, clip at bounds, else resample [default = False]
        smap: map instance, to sample the quantity in parallel [default = None]
        npts: number of sample points [default = 10000]

    Returns:
        sampled statistical quantity, for the specified axis, reduced to a float
        """
        #XXX: return what? "energy and solution?" reduced?
        if 'map' in kwds and 'smap' in kwds:
            msg = "__call__() can either accept 'smap' or 'map', not both"
            raise TypeError(msg)
        tight = kwds.pop('tight', True) #XXX: better False?
        archive = kwds.pop('archive', None)
        if archive is None: archive = dict_archive()
        elif type(archive) in (str, (u''.__class__)):
            archive = read(archive, type=file_archive)
        smap = kwds.pop('smap', kwds.pop('map', None))
        if smap is None: smap = map #NOTE: was _ThreadPool.map w/ join
        npts = kwds.get('npts', None)
        if npts is None: kwds.pop('npts', None)
        fobj = cached(archive=archive)(self.objective) #XXX: bad idea?
        self._pts = fobj.__cache__() #XXX: also bad idea? include from *_bounds?
        if tight:
            from mystic.constraints import and_, boundsconstrain
            bounds = boundsconstrain(self.lb, self.ub) #XXX: symbolic?
            constraint = and_(self.constraint, bounds, onfail=bounds)
        else:
            constraint = self.constraint
        objective = lambda rv: fobj(constraint(rv), axis=axis) # penalty?
        _pts = len(self._pts)
        pts = npts - _pts
        self._invalid, invalid = float('nan'), self._invalid
        if pts > 0: # need to sample some new points
            kwds['npts'] = pts
            s = random_samples(self.lb, self.ub, **kwds).T
            if _pts: #NOTE: s = [(...),(...)] or [...]
                s = list(self._pts.values()) + list(smap(objective, s))
            else:
                s = list(smap(objective, s))
        elif pts == 0:
            s = list(self._pts.values())
        else: # randomly choose points from archive
            import numpy as np
            s = np.random.choice(range(_pts), size=npts, replace=False)
            s = np.array(list(self._pts.values()))[s].tolist()
        self._invalid = invalid
        '''
        print(s[-1])
        # mark error (nan/inf) in keys with nan in values
        import numpy as np
        keys = list(self._pts.keys())
        bad = np.where(np.isfinite(np.array(keys).sum(axis=-1)) == False)[0]
        for k in bad:
            self._pts[keys[k]] = float('nan')
        # convert errors in values to nan
        bad = np.isfinite(s)
        multi = bad.ndim > 1
        if multi:
            bad = bad.all(axis=-1)
        bad = np.where(bad == False)[0]
        for k in bad:
            s[k] = tuple(float('nan') for i in s[k]) if multi else float('nan')
        '''
        # calculate expected value
        if axis is None and self.axes is not None: # apply per axis
            if reducer is None: #XXX: nanreducer?
                import numpy as np
                return tuple(np.nanmean(s, axis=0).tolist())
            #XXX: better tuple(reducer(s, axis=0).tolist()) if numpy ufunc?
            return tuple(reducer(si) for si in zip(*s))
        if axis is None:
            s = tuple(s)
        else:
            s = list(zip(*s))[axis]
        if reducer is None:
            import numpy as np
            return np.nanmean(s).tolist()
        return reducer(s)

    # --- solve ---
    def solve(self, objective, **kwds): #NOTE: single axis only
        """solve (in measure space) for bound on given objective

    Input:
        objective: cost function of the form y = objective(x)

    Additional Input:
        solver: mystic.solver instance [default: DifferentialEvolutionSolver2]
        npop: population size [default: None]
        id: a unique identifier for the solver [default: None]
        nested: mystic.solver instance [default: None], for ensemble solvers
        x0: initial parameter guess [default: use RandomInitialPoints]
        maxiter: max number of iterations [default: defined in solver]
        maxfun: max number of objective evaluations [default: defined in solver]
        evalmon: mystic.monitor instance [default: Monitor], for evaluations
        stepmon: mystic.monitor instance [default: Monitor], for iterations
        map: pathos map instance for solver.SetMapper [default: None]
        save: iteration frequency to save solver [default: None]
        opts: dict of configuration options for solver.Solve [default: {}]

    Returns:
        solver instance, after Solve has been called
        """
        #NOTE: kwds(penalty, stop) undocumented
        penalty = kwds.pop('penalty', None) # special case 'and(penalty)'
        stop = kwds.pop('stop', None) # special case 'Or(stop)'
        k = self.kwds.copy(); k.update(kwds) # overrides self.kwds
        kwds.update(k) #FIXME: good idea??? [bad in parallel???]
        lb, ub = self.lb, self.ub
        solver = kwds.get('solver', DifferentialEvolutionSolver2)
        npop = kwds.get('npop', None)
        if npop is not None or kwds.get('NP', None) is not None:
            solver = solver(len(lb),npop)
        else:
            solver = solver(len(lb))
        solver.id = kwds.pop('id', None)
        nested = kwds.get('nested', None)
        x0 = kwds.get('x0', None)
        if nested is not None: # Buckshot/Sparsity
            solver.SetNestedSolver(nested)
        else: # DiffEv/Nelder/Powell
            if x0 is None: solver.SetRandomInitialPoints(min=lb,max=ub)
            else: solver.SetInitialPoints(x0)
        save = kwds.get('save', None)
        if save is not None:
            solver.SetSaveFrequency(save, 'Solver.pkl') #FIXME: set name
        mapper = kwds.get('map', None)
        if mapper is not None:
            solver.SetMapper(mapper) #NOTE: not Nelder/Powell
        maxiter = kwds.get('maxiter', None)
        maxfun = kwds.get('maxfun', None)
        solver.SetEvaluationLimits(maxiter,maxfun)
        evalmon = kwds.get('evalmon', None)
        evalmon = Monitor() if evalmon is None else evalmon
        solver.SetEvaluationMonitor(evalmon[:0])
        stepmon = kwds.get('stepmon', None)
        stepmon = Monitor() if stepmon is None else stepmon[:0]
        solver.SetGenerationMonitor(stepmon)
        solver.SetStrictRanges(min=lb,max=ub)#,tight=True) #XXX: tight?
        solver.SetConstraints(self.constraint)
        if penalty is not None: # add the special-case penalty
            from mystic.coupler import and_
            penalty = and_(self.penalty, penalty)
        else:
            penalty = self.penalty
        solver.SetPenalty(penalty)
        opts = kwds.get('opts', {}) #XXX: copy is necessary?
        if stop is not None: # add the special-case termination
            term = opts.get('termination', solver._termination)
            from mystic.termination import Or
            opts['termination'] = Or(stop, term)
        # solve
        solver.Solve(objective, **opts)
        if mapper is not None:
            mapper.close()
            mapper.join()
            mapper.clear() #NOTE: if used, then shut down pool
        #NOTE: debugging code
        #print("solved: %s" % solver.Solution())
        #func_bound = solver.bestEnergy
        #func_evals = solver.evaluations
        #from mystic.munge import write_support_file
        #write_support_file(solver._stepmon)
        #print("func_bound: %s" % func_bound) #NOTE: may be inverted
        #print("func_evals: %s" % func_evals)
        return solver


class ExpectedValue(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate expected value of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the expected value for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_expect if samples, else expect
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            if axis is None and self.axes is not None:
                return (self._invalid,) * (self.axes or 1) #XXX:?
            return self._invalid
        # get expected value
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.expect(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_expect(m, self.samples, map=self.map) for m in model)
        # else, get expected value for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            return c.expect(model)
        return c.sampled_expect(model, self.samples, map=self.map)


class Variance(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate expected variance of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the expected variance for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_variance if samples, else expect_var
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            if axis is None and self.axes is not None:
                return (self._invalid,) * (self.axes or 1) #XXX:?
            return self._invalid
        # get expected variance
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.expect_var(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_variance(m, self.samples, map=self.map) for m in model)
        # else, get expected variance for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            return c.expect_var(model)
        return c.sampled_variance(model, self.samples, map=self.map)


class MaximumValue(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate maximum value of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the maximum value for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_maximum if samples, else ess_maximum
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            if axis is None and self.axes is not None:
                return (self._invalid,) * (self.axes or 1) #XXX:?
            return self._invalid
        # get maximum value
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.ess_maximum(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_maximum(m, self.samples, map=self.map) for m in model)
        # else, get maximum value for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            from mystic.math.measures import ess_maximum #TODO: c.ess_maximum
            return ess_maximum(model, c.positions, c.weights)
        return c.sampled_maximum(model, self.samples, map=self.map)


class MinimumValue(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate minimum value of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the minimum value for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_minimum if samples, else ess_minimum
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            if axis is None and self.axes is not None:
                return (self._invalid,) * (self.axes or 1) #XXX:?
            return self._invalid
        # get minimum value
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.ess_minimum(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_minimum(m, self.samples, map=self.map) for m in model)
        # else, get minimum value for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            from mystic.math.measures import ess_minimum #TODO: c.ess_minimum
            return ess_minimum(model, c.positions, c.weights)
        return c.sampled_minimum(model, self.samples, map=self.map)


class ValueAtRisk(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate value at risk of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the value at risk for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_ptp if samples, else ess_ptp
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            if axis is None and self.axes is not None:
                return (self._invalid,) * (self.axes or 1) #XXX:?
            return self._invalid
        # get value at risk
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.ess_ptp(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_ptp(m, self.samples, map=self.map) for m in model)
        # else, get value at risk for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            from mystic.math.measures import ess_ptp #TODO: c.ess_ptp
            return ess_ptp(model, c.positions, c.weights)
        return c.sampled_ptp(model, self.samples, map=self.map)


class ProbOfFailure(BaseOUQ):

    def objective(self, rv, axis=None, iter=True):
        """calculate probability of failure for model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)
        iter: bool, if True, calculate per axis, else calculate for all axes

    Returns:
        the probability of failure for the specified axis (or all axes)

    NOTE:
        respects constraints on input parameters and product measure
        model is a function returning a boolean (True for success)

    NOTE:
        for product_measure, use sampled_pof(model) if samples, else pof(model)
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            if axis is None and self.axes is not None:
                return (self._invalid,) * (self.axes or 1) #XXX:?
            return self._invalid
        # get probability of failure
        if iter and axis is None and self.axes is not None:
            model = (lambda x: self.model(x,axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.pof(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_pof(m, self.samples, map=self.map) for m in model)
        # else, get probability of failure for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            return c.pof(model)
        return c.sampled_pof(model, self.samples, map=self.map)




#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

def objective(x):
    x0,x1 = x
    return x0**2 + (x1 - 1)**2

from mystic.bounds import Bounds
bounds = Bounds(-1,1,n=2)
# with penalty='penalty' applied, solution is:
xs, xs_ = [(0.5)**.5, 0.5], [-(0.5)**.5, 0.5]
ys = 0.75

from mystic.symbolic import generate_constraint, generate_solvers, solve
from mystic.symbolic import generate_penalty, generate_conditions

equations = """
x1 - x0**2 = 0.0
"""
cf = generate_constraint(generate_solvers(solve(equations, target='x1')))
pf = generate_penalty(generate_conditions(equations), k=1e12)



if __name__ == '__main__':

    from mystic.solvers import diffev2
    from mystic.math import almostEqual

    result = diffev2(objective, x0=bounds, bounds=bounds, constraints=cf, npop=40, disp=False, full_output=True)

    assert almostEqual(result[0], xs, tol=1e-2) \
        or almostEqual(result[0], xs_, tol=1e-2)
    assert almostEqual(result[1], ys, rel=1e-2)



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2020-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"Hyperparameter optimization"

from sklearn.svm import SVR
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split as tts
iris = load_iris()
X_train, X_test, y_train, y_test = tts(iris.data, iris.target, random_state=1)


def objective(x):
    estimator = SVR(kernel='linear', C=x[0])
    estimator.fit(X_train, y_train)
    return 1-estimator.score(X_test, y_test)


from mystic.bounds import Bounds
bounds = Bounds(0,10)
# for the given split, the solution is, roughly:
xs = [0.01213213]
ys =  0.08483955


if __name__ == '__main__':

    from mystic.solvers import diffev2
    from mystic.math import almostEqual
    from mystic.monitors import VerboseMonitor
    mon = VerboseMonitor(10)

    result = diffev2(objective, x0=bounds, bounds=bounds, npop=40, ftol=1e-8, gtol=75, disp=False, full_output=True, itermon=mon)

    assert almostEqual(result[0], xs, rel=1e-2)
    assert almostEqual(result[1], ys, rel=1e-2)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2019-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
samplers: optimizer-guided directed sampling
"""
__all__  = ['LatticeSampler','BuckshotSampler','SparsitySampler',\
            'ResidualSampler','MixedSampler']

from mystic.abstract_sampler import AbstractSampler


class LatticeSampler(AbstractSampler):
    """
optimizer-directed sampling starting at the centers of N grid points
    """
    def _init_solver(self, **kwd):
        """initialize the ensemble solver"""
        from mystic.ensemble import LatticeSolver
        s = LatticeSolver(len(self._bounds), nbins=self._npts)
        s.SetStrictRanges(*zip(*self._bounds), **kwd)
        s.SetObjective(self._model)
        return s


class BuckshotSampler(AbstractSampler):
    """
optimizer-directed sampling starting at N randomly sampled points
    """
    def _init_solver(self, **kwd):
        """initialize the ensemble solver"""
        from mystic.ensemble import BuckshotSolver
        s = BuckshotSolver(len(self._bounds), npts=self._npts)
        s.SetStrictRanges(*zip(*self._bounds), **kwd)
        s.SetObjective(self._model)
        return s


class SparsitySampler(AbstractSampler):
    """
optimizer-directed sampling starting at N points sampled in sparse reigons
    """
    def __init__(self, bounds, model, npts=None, **kwds):
        self._rtol = kwds.pop('rtol', None)
        super(SparsitySampler, self).__init__(bounds, model, npts, **kwds)
        return
    __init__.__doc__ = AbstractSampler.__init__.__doc__
    def _init_solver(self, **kwd):
        """initialize the ensemble solver"""
        from mystic.ensemble import SparsitySolver
        s = SparsitySolver(len(self._bounds), npts=self._npts, rtol=self._rtol)
        s.SetStrictRanges(*zip(*self._bounds), **kwd)
        s.SetObjective(self._model)
        return s


class ResidualSampler(AbstractSampler):
    """
optimizer-directed sampling starting at N points sampled near largest misfit
    """
    def __init__(self, bounds, model, npts=None, **kwds):
        self._mtol = kwds.pop('mtol', None)
        super(ResidualSampler, self).__init__(bounds, model, npts, **kwds)
        return
    __init__.__doc__ = AbstractSampler.__init__.__doc__
    def _init_solver(self, **kwd):
        """initialize the ensemble solver"""
        from mystic.ensemble import ResidualSolver
        s = ResidualSolver(len(self._bounds), npts=self._npts, mtol=self._mtol, func=self._model) #FIXME: TEST (mon?)
        s.SetStrictRanges(*zip(*self._bounds), **kwd)
        s.SetObjective(self._model)
        return s


class MixedSampler(AbstractSampler):
    """
optimizer-directed sampling using N points from a mixture of ensemble solvers
    """
    def _init_solver(self, **kwd):
        """initialize the ensemble solver"""
        from mystic.ensemble import MixedSolver
        s = MixedSolver(len(self._bounds), samp=self._npts)
        s.SetStrictRanges(*zip(*self._bounds), **kwd)
        s.SetObjective(self._model)
        return s




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example:
    - Minimize Rosenbrock's Function with Powell's method.
    - Dynamic print of parameter convergence to function minimum.

Demonstrates:
    - standard models
    - minimal solver interface
    - parameter trajectories using callback
"""

# Powell's Directonal solver
from mystic.solvers import fmin_powell

# Rosenbrock function
from mystic.models import rosen

iter = 0
# plot the parameter trajectories
def print_params(params):
    global iter
    from numpy import asarray
    print("Generation %d has best fit parameters: %s" % (iter,asarray(params)))
    iter += 1
    return


if __name__ == '__main__':

    # initial guess
    x0 = [0.8,1.2,0.7]
    print_params(x0)

    # use Powell's method to minimize the Rosenbrock function
    solution = fmin_powell(rosen,x0,disp=1,callback=print_params,handler=True)
    print(solution)

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Patrick Hung (patrickh @caltech)
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Copyright (c) 2010-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Factories that provide termination conditions for a mystic.solver
"""

import numpy
np = numpy
abs = numpy.absolute
inf = numpy.inf
nan = numpy.nan
null = ""
_type = type #NOTE: save builtin type
import mystic.collapse as ct #XXX: avoid if move Collapse* to collapse
from mystic.math.distance import Lnorm #XXX: avoid this and following?
from mystic._scipy060optimize import approx_fprime, _epsilon

# a module level singleton
EARLYEXIT = 0

# termination condition interrogator functions
#FIXME FIXME: assumes NO DUPLICATE TYPES of termination conditions
def state(condition):
    '''get state (dict of kwds) used to create termination condition'''
    #NOTE: keys are termination name; values are termination state
    _state = {}
    for term in iter(condition) if isinstance(condition, tuple) else iter((condition,)):
        termdoc = term.__doc__# or 'MISSING'
        if termdoc is None: #HACK
            import warnings
            warnings.warn('Collapse termination mishandled')
            pass #FIXME: HACK: shouldn't be missing (due to ensemble + Mapper)
        elif not termdoc.split(None,1)[-1].startswith('with '):# Or, And, ...
            _state.update(state(term))
        else:
            kind,kwds = termdoc.split(' with ', 1)
           #_state[kind] = eval(kwds)
            _state[termdoc] = eval(kwds)
    return _state

def type(condition):
    '''get object that generated the given termination instance'''
    if isinstance(condition, _type(lambda :None)): #XXX: type of term conditions
        try:
            import importlib
            module = importlib.import_module(condition.__module__)
        except ImportError:
            module = __import__(condition.__module__, globals(), locals(), ['object'], -1)
        return getattr(module, condition.__name__[1:]) #XXX: start w/ _
    # otherwise, just figure it's a class or standard object
    return _type(condition)


# Factories that extend termination conditions
class When(tuple):
  """provide a termination condition with more reporting options.

Terminates when the given condition is satisfied."""
  def __new__(self, arg):
    """
Takes a termination condition:
    arg    -- termination condition

Example:
    >>> from mystic.termination import When, VTR
    >>> term = When( VTR() )
    >>> term(solver)  # where solver is a mystic.solver instance
    """
    if isinstance(arg, tuple) and len(arg) == 1: arg = arg[0] # for pickling
    #XXX: need better filter on inputs
    if getattr(arg, '__module__', None) != self.__module__:
      raise TypeError("'%s' object is not a condition" % arg.__class__.__name__)
    if not getattr(arg, '__len__', None): arg = [arg]
    return tuple.__new__(self, arg)

  def __call__(self, solver, info=False):
    """check if the termination conditions are satisfied.

Inputs:
    solver -- the solver instance

Additional Inputs:
    info   -- if True, return information about the satisfied conditions"""
    # return the unsatisfied conditions
    if info == 'not':
      return tuple(set([f for f in self if f not in self(solver, 'self')]))
    # do some filtering...
    stop = {}
    [stop.update({f : f(solver, info)}) for f in self]
    _all = all(stop.values())
    # return T/F if the conditions are met
    if not info: return _all
    # return the satisfied conditions
    if info == 'self': return tuple(set(stop.keys())) if _all else ()
    # return info about the satisfied conditions
    return "; ".join(set("; ".join(stop.values()).split("; "))) if _all else ""

  def __repr__(self):
    return "When(%s)" % str(self[0])

class And(When):
  """couple termination conditions with "and".

Terminates when all given conditions are satisfied."""
  def __new__(self, *args):
    """
Takes one or more termination conditions:
    args   -- tuple of termination conditions

Example:
    >>> from mystic.termination import And, VTR, ChangeOverGeneration
    >>> term = And( VTR(), ChangeOverGeneration() )
    >>> term(solver)  # where solver is a mystic.solver instance
    """
    if isinstance(args, tuple) and len(args) == 1: args = args[0] # for pickling
    #XXX: need better filter on inputs
    if not getattr(args, '__len__', None): args = [args]
    #XXX: check if every arg in args has __module__ == self.__module__ ?
    return tuple.__new__(self, args)

  def __repr__(self):
    return "And%s" % str(tuple([f for f in self]))


class Or(When):
  """couple termination conditions with "or".

Terminates when any of the given conditions are satisfied."""
  def __new__(self, *args):
    """
Takes one or more termination conditions:
    args   -- tuple of termination conditions

Example:
    >>> from mystic.termination import Or, VTR, ChangeOverGeneration
    >>> term = Or( VTR(), ChangeOverGeneration() )
    >>> term(solver)  # where solver is a mystic.solver instance
    """
    if isinstance(args, tuple) and len(args) == 1: args = args[0] # for pickling
    #XXX: need better filter on inputs
    if not getattr(args, '__len__', None): args = [args]
    #XXX: check if every arg in args has __module__ == self.__module__ ?
    return tuple.__new__(self, args)

  def __call__(self, solver, info=False):
    """check if the termination conditions are satisfied.

Inputs:
    solver -- the solver instance

Additional Inputs:
    info   -- if True, return information about the satisfied conditions"""
    # return the unsatisfied conditions
    if info == 'not':
      return tuple(set([f for f in self if f not in self(solver, 'self')]))
    # do some filtering...
    stop = {}
    [stop.update({f : f(solver, info)}) for f in self]
    _any = any(stop.values())
    # return T/F if the conditions are met
    if not info: return _any
    [stop.pop(cond) for (cond,met) in tuple(stop.items()) if not met]
    # return the satisfied conditions
    if info == 'self': return tuple(set(stop.keys()))
    # return info about the satisfied conditions
    return "; ".join(set("; ".join(stop.values()).split("; ")))

  def __repr__(self):
    return "Or%s" % str(tuple([f for f in self]))


# Factories that give termination conditions
#FIXME: the following should be refactored into classes
def VTR(tolerance=0.005, target=0.0):
    """cost of last iteration is < tolerance from target:

``abs(cost[-1] - target) <= tolerance``
"""
    doc = "VTR with %s" % {'tolerance':tolerance, 'target':target}
    def _VTR(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        if not len(hist): return info(null)
        if abs(hist[-1] - target) <= tolerance: return info(doc)
        return info(null)
   #_VTR.__doc__ = "%s(**%s)" % tuple(doc.split(" with "))
    _VTR.__doc__ = doc
    return _VTR

def ChangeOverGeneration(tolerance=1e-6, generations=30):
    """change in cost is < tolerance over a number of generations:

``cost[-g] - cost[-1] <= tolerance``, with ``g=generations``
"""
    doc = "ChangeOverGeneration with %s" % {'tolerance':tolerance,
                                            'generations':generations}
    def _ChangeOverGeneration(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        gens = 0 if generations is None else int(generations) #NOTE: floor
        if lg <= gens: return info(null)
        if (hist[-gens]-hist[-1]) <= tolerance: return info(doc)
        if (hist[-gens] == hist[-1]): return info(doc)
        return info(null)
    _ChangeOverGeneration.__doc__ = doc
    return _ChangeOverGeneration

def NormalizedChangeOverGeneration(tolerance=1e-4, generations=10):
    """normalized change in cost is < tolerance over number of generations:

``(cost[-g] - cost[-1]) / 0.5*(abs(cost[-g]) + abs(cost[-1])) <= tolerance``
"""
    eta = 1e-20
    doc = "NormalizedChangeOverGeneration with %s" % {'tolerance':tolerance,
                                                      'generations':generations}
    def _NormalizedChangeOverGeneration(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        gens = 0 if generations is None else int(generations) #NOTE: floor
        if lg <= gens: return info(null)
        if (hist[-gens] == hist[-1]): return info(doc)
        diff = tolerance*(abs(hist[-gens])+abs(hist[-1])) + eta
        if 2.0*(hist[-gens]-hist[-1]) <= diff: return info(doc)
        return info(null)
    _NormalizedChangeOverGeneration.__doc__ = doc
    return _NormalizedChangeOverGeneration
              
def CandidateRelativeTolerance(xtol=1e-4, ftol=1e-4):
    """absolute difference in candidates is < tolerance:

``abs(xi-x0) <= xtol`` & ``abs(fi-f0) <= ftol``, with ``x=params`` & ``f=cost``
"""
    #NOTE: this termination expects nPop > 1
    doc = "CandidateRelativeTolerance with %s" % {'xtol':xtol, 'ftol':ftol}
    def _CandidateRelativeTolerance(inst, info=False):
        sim = numpy.array(inst.population)
        fsim = numpy.array(inst.popEnergy)
        if not len(fsim[1:]):
            warn = "Warning: Invalid termination condition (nPop < 2)"
            print(warn)
            return warn
        #   raise ValueError, "Invalid termination condition (nPop < 2)"
        if info: info = lambda x:x
        else: info = bool
        #FIXME: abs(inf - inf) will raise a warning...
        errdict = numpy.seterr(invalid='ignore') #FIXME: turn off warning 
        answer = max(numpy.ravel(abs(sim[1:]-sim[0]))) <= xtol
        answer = answer and max(abs(fsim[0]-fsim[1:])) <= ftol
        numpy.seterr(invalid=errdict['invalid']) #FIXME: turn on warnings
        if answer: return info(doc)
        return info(null)
    _CandidateRelativeTolerance.__doc__ = doc
    return _CandidateRelativeTolerance

def SolutionImprovement(tolerance=1e-5):  
    """sum of change in each parameter is < tolerance:

``sum(abs(last_params - current_params)) <= tolerance``
"""
    doc = "SolutionImprovement with %s" % {'tolerance':tolerance}
    def _SolutionImprovement(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        best = numpy.array(inst.bestSolution)
        trial = numpy.array(inst.trialSolution)
        update = abs(best - trial) #XXX: if inf - inf ?
        answer = numpy.add.reduce(update.T)
        if isinstance(answer, numpy.ndarray): # if trialPop, take 'best' answer
            answer = max(answer)              #XXX: is this 'best' or 'worst'?
        answer = answer <= tolerance
        if answer: return info(doc)
        return info(null)
    _SolutionImprovement.__doc__ = doc
    return _SolutionImprovement

def NormalizedCostTarget(fval=None, tolerance=1e-6, generations=30):
    """normalized absolute difference from given cost value is < tolerance:
(if fval is not provided, then terminate when no improvement over g iterations)

``abs(cost[-1] - fval)/fval <= tolerance`` or ``(cost[-1] - cost[-g]) = 0``
"""
    #NOTE: modified from original behavior
    #  original --> if generations: then return cost[-g] - cost[-1] < 0
    #           --> else: return fval != 0 and abs((best - fval)/fval) < tol
    doc = "NormalizedCostTarget with %s" % {'fval':fval, 'tolerance':tolerance,
                                            'generations':generations}
    def _NormalizedCostTarget(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        gens = 0 if generations is None else int(generations) #NOTE: floor
        if gens and fval is None:
            #XXX: throws error when hist is shorter than generations ?
            if lg > gens and ((hist[-gens]-hist[-1]) <= 0 or \
                                     (hist[-gens] == hist[-1])):
                return info(doc)
            return info(null)
        if not gens and fval is None: return info(doc)
        if abs(hist[-1]-fval) <= abs(tolerance * fval): return info(doc)
        return info(null)
    _NormalizedCostTarget.__doc__ = doc
    return _NormalizedCostTarget

def VTRChangeOverGeneration(ftol=0.005, gtol=1e-6, generations=30, target=0.0):
    """change in cost is < gtol over a number of generations,
or cost of last iteration is < ftol from target:

``cost[-g] - cost[-1] <= gtol`` or ``abs(cost[-1] - target) <= ftol``
"""
    doc = "VTRChangeOverGeneration with %s" % {'ftol':ftol, 'gtol':gtol,
                                               'generations':generations,
                                               'target':target}
    def _VTRChangeOverGeneration(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        gens = 0 if generations is None else int(generations) #NOTE: floor
        #XXX: throws error when hist is shorter than generations ?
        if (lg > gens and ((hist[-gens]-hist[-1]) <= gtol or \
                                  (hist[-gens] == hist[-1]))) or \
           ( abs(hist[-1] - target) <= ftol ): return info(doc)
        return info(null)
    _VTRChangeOverGeneration.__doc__ = doc
    return _VTRChangeOverGeneration

def PopulationSpread(tolerance=1e-6):
    """normalized absolute deviation from best candidate is < tolerance:

``abs(params - params[0]) <= tolerance``
"""
    doc = "PopulationSpread with %s" % {'tolerance':tolerance}
    def _PopulationSpread(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        sim = numpy.array(inst.population)
        #if not len(sim[1:]):
        #    warn = "Warning: Invalid termination condition (nPop < 2)"
        #    print(warn)
        #    return warn
        if numpy.all(abs(sim - sim[0]) <= abs(tolerance * sim[0])): return info(doc)
        return info(null)
    _PopulationSpread.__doc__ = doc
    return _PopulationSpread

def GradientNormTolerance(tolerance=1e-5, norm=inf): 
    """gradient norm is < tolerance, given user-supplied norm:

``sum( abs(gradient)**norm )**(1.0/norm) <= tolerance``
"""
    doc = "GradientNormTolerance with %s" % {'tolerance':tolerance, 'norm':norm}
    def _GradientNormTolerance(inst, info=False):
        grad = getattr(inst, 'gradient', [None])[-1]
        if grad is None:
            soln = inst.bestSolution
            cost = inst._cost[1]
            grad = approx_fprime(soln, cost, _epsilon)
           #warn = "Warning: using approximate gradient"
           #print(warn)
           #return warn
        if info: info = lambda x:x
        else: info = bool
        gnorm = Lnorm(grad, p=norm, axis=0)
        if gnorm <= tolerance: return info(doc)
        return info(null)
    _GradientNormTolerance.__doc__ = doc
    return _GradientNormTolerance

def EvaluationLimits(generations=None, evaluations=None):
    """number of iterations is > generations,
or number of function calls is > evaluations:

``iterations >= generations`` or ``fcalls >= evaluations``
"""
    #NOTE: default settings use solver defaults (_maxfun and _maxiter)
    doc = "EvaluationLimits with %s" % {'generations':generations, \
                                        'evaluations':evaluations}
    maxfun = [evaluations]
    maxiter = [generations]
    if maxfun[0] is None: maxfun[0] = inf 
    if maxiter[0] is None: maxiter[0] = inf
    def _EvaluationLimits(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        gens = inst.generations
        eval = inst._fcalls[0]
        if (eval >= maxfun[0]) or (gens >= maxiter[0]): return info(doc)
        return info(null)
   #_EvaluationLimits.__doc__ = "%s(**%s)" % tuple(doc.split(" with "))
    _EvaluationLimits.__doc__ = doc
    return _EvaluationLimits

# process_time -- CPU execution time (process-wide)
# perf_counter -- wall time since starting __main__ (system-wide)
# time -- time since Epoch (global)
def TimeLimits(seconds=86400, system=None): #NOTE: 24 hours
    """elapsed time is > seconds, using an independent global counter:
(to guarantee monotonicity, use system=True for a system-wide counter,
 or system=False for a counter based on current-process execution time)

``time >= seconds``
"""
    doc = "TimeLimits with %s" % {'seconds':seconds, 'system':system}
    import time
    if system is None: timer = time.time
    elif system: timer = time.perf_counter
    else: timer = time.process_time
    start = [timer()]
    def _reset():
        start[0] = timer()
    delta = [getattr(seconds, 'total_seconds', seconds.__abs__)()]
    if delta[0] is None: delta[0] = inf
    def _TimeLimits(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        if (timer() - start[0]) >= delta[0]: return info(doc)
        return info(null)
    _TimeLimits.__doc__ = doc
    _TimeLimits.reset = _reset
    return _TimeLimits

def SolverInterrupt(): #XXX: enable = True ?
    """handler is enabled and interrupt is given:

``_EARLYEXIT == True``
"""
    doc = "SolverInterrupt with %s" % {}
    def _SolverInterrupt(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        if inst._EARLYEXIT: return info(doc)
        return info(null)
    _SolverInterrupt.__doc__ = doc
    return _SolverInterrupt

##### parameter collapse conditions #####
def CollapseWeight(tolerance=0.005, generations=50, mask=None, **kwds):
    """value of weights are < tolerance over a number of generations,
where mask is (row,column) indices of the selected weights:

``bool(collapse_weight(monitor, **kwds))``
"""
    _kwds = {'tolerance':tolerance, 'generations':generations, 'mask':mask}
    kwds.update(_kwds)
    doc = "CollapseWeight with %s" % kwds #XXX: better kwds or _kwds?
    def _CollapseWeight(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        if lg <= generations: return info(null)
        #XXX: might want to log/utilize *where* collapse happens...
#       if collapse_weight(inst._stepmon, **kwds): return info(doc)
        collapsed = ct.collapse_weight(inst._stepmon, **kwds)
        if collapsed: return info(doc + ' at %s' % str(collapsed))
        # otherwise bail out
        return info(null) 
    _CollapseWeight.__doc__ = doc
    return _CollapseWeight

def CollapsePosition(tolerance=0.005, generations=50, mask=None, **kwds):
    """max(pairwise(positions)) < tolerance over a number of generations,
where (measures,indices) are (row,column) indices of selected positions:

``bool(collapse_position(monitor, **kwds))``
"""
    _kwds = {'tolerance':tolerance, 'generations':generations, 'mask':mask}
    kwds.update(_kwds)
    doc = "CollapsePosition with %s" % kwds #XXX: better kwds or _kwds
    def _CollapsePosition(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        if lg <= generations: return info(null)
        #XXX: might want to log/utilize *where* collapse happens...
#       if collapse_weight(inst._stepmon, **kwds): return info(doc)
        collapsed = ct.collapse_position(inst._stepmon, **kwds)
        if collapsed: return info(doc + ' at %s' % str(collapsed))
        # otherwise bail out
        return info(null) 
    _CollapsePosition.__doc__ = doc
    return _CollapsePosition

def CollapseAt(target=None, tolerance=1e-4, generations=50, mask=None):
    """change(x[i]) is < tolerance over a number of generations,
where target can be a single value or a list of values of x length,
change(x[i]) = max(x[i]) - min(x[i]) if target=None else abs(x[i] - target),
and mask is column indices of selected params:

``bool(collapse_at(monitor, **kwds))``
"""
    kwds = {'tolerance':tolerance, 'generations':generations,
            'target':target, 'mask':mask}
    doc = "CollapseAt with %s" % kwds
    def _CollapseAt(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        if lg <= generations: return info(null)
        #XXX: might want to log/utilize *where* collapse happens...
#       if ct.collapse_at(inst._stepmon, **kwds): return info(doc)
        collapsed = ct.collapse_at(inst._stepmon, **kwds)
        if collapsed: return info(doc + ' at %s' % str(collapsed))
        # otherwise bail out
        return info(null) 
    _CollapseAt.__doc__ = doc
    return _CollapseAt

def CollapseAs(offset=False, tolerance=1e-4, generations=50, mask=None):
    """max(pairwise(x)) is < tolerance over a number of generations,
and mask is column indices of selected params:

``bool(collapse_as(monitor, **kwds))``
"""
    kwds = {'tolerance':tolerance, 'generations':generations,
            'offset':offset, 'mask':mask}
    doc = "CollapseAs with %s" % kwds
    def _CollapseAs(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        if lg <= generations: return info(null)
        #XXX: might want to log/utilize *where* collapse happens...
#       if ct.collapse_as(inst._stepmon, **kwds): return info(doc)
        collapsed = ct.collapse_as(inst._stepmon, **kwds)
        if collapsed: return info(doc + ' at %s' % str(collapsed))
        # otherwise bail out
        return info(null) 
    _CollapseAs.__doc__ = doc
    return _CollapseAs

##### bounds collapse conditions #####
def CollapseCost(clip=False, limit=1.0, samples=50, mask=None):
    """cost(x) - min(cost) is >= limit for all samples within an interval,
where if clip is True, then clip beyond the space sampled the optimizer,
and mask is a dict of {index:bounds} where bounds are provided as an
interval (min,max), or a list of intervals:

``bool(collapse_cost(monitor, **kwds))``
"""
    kwds = {'limit':limit, 'samples':samples,
            'clip':clip, 'mask':mask}
    doc = "CollapseCost with %s" % kwds
    def _CollapseCost(inst, info=False):
        if info: info = lambda x:x
        else: info = bool
        hist = inst.energy_history
        lg = len(hist)
        if not lg: return info(null)
        if lg <= samples: return info(null)
        #XXX: mask = interval_overlap(mask, solver_bounds(inst))?
        #XXX: might want to log/utilize *where* collapse happens...
#       if collapse_cost(inst._stepmon, **kwds): return info(doc)
        collapsed = ct.collapse_cost(inst._stepmon, **kwds)
        if collapsed: return info(doc + ' at %s' % str(collapsed))
        # otherwise bail out
        return info(null)
    _CollapseCost.__doc__ = doc
    return _CollapseCost

# end of file



#!/usr/bin/env python
# 
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2024-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
noisy emulator:
* input parameters (a_alpha, c_alpha, a_steel)
  with solved values at:
    [2.9306538, 4.6817646, 3.6026807]
  measured error of:
    [1e-4, 1e-4, 1e-4]
  bounds at:
    [(0.95 * i, 1.05 * i) for i in solved]
  yielding a minimum at:
    a_beta = 3.233392
  with a measured error of:
    a_beta_error = 1e-4
  and gaussian noise on the input parameters of:
    sigma = 0.001
'''
from itertools import repeat
from mystic.monitors import LoggingMonitor
from emulators import cost3, x3, bounds3
from noisy import noisy

try: # parallel maps
    from pathos.maps import Map
    from pathos.pools import ProcessPool, ThreadPool, SerialPool
    cmap = Map(SerialPool) #ProcessPool
except ImportError:
    cmap = map

# use a logging monitor to track cost function evaluations
monitor3 = LoggingMonitor(1, filename='truth.txt', label='truth')

# generate an objective with gaussian noise on the inputs
noisycost3 = lambda x: cost3(noisy(x, sigma=.001))

# generate a tracked cost function
def trackcost3(x):
  fx = noisycost3(x)
  monitor3(x, fx)
  return fx

# average the cost over the number of cost evaluations per objective call
avecost3 = lambda x, n=1: sum(cmap(trackcost3, repeat(x, n)))/n

# set the number of cost evaluations per objective call
ave10cost3 = lambda x: avecost3(x, n=10)



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.tools import list_or_tuple_or_ndarray as sequence
from mystic.tools import isNull

import sys

# generalized history reader

def read_history(source, iter=False):
    """read parameter history and cost history from the given source

source is a monitor, logfile, support file, solver restart file, dataset, etc
iter is a bool, where if True, `iter` is also returned

Returns (iter, params, cost) if iter is True, else returns (params, cost), where
    - iter is a list of tuples of ints (iteration, id), where id may be None
    - params is input to cost; a list of lists of floats
    - cost is output of cost; a list of floats, or a list of tuples of floats
    """
    monitor = solver = data = arxiv = False
    from mystic.monitors import Monitor, Null
    from mystic.abstract_solver import AbstractSolver
    from mystic.math.legacydata import dataset
    from klepto.archives import archive, cache
    import io
    file = io.IOBase
    if isinstance(source, file):
        return read_history(source.name, iter)
    if isinstance(source, str):
        import re
        source = re.sub(r'\.py*.$', '', source)  # strip off .py* extension
    elif isinstance(source, Monitor):
        monitor = True
    elif isinstance(source, dataset):
        data = True
    elif isinstance(source, cache):
        source = source.__archive__
        arxiv = True
    elif isinstance(source, archive):
        arxiv = True
    elif isinstance(source, AbstractSolver):
        solver = True
    elif isinstance(source, Null): #XXX: or source.x, source.y or Error?
        return ([],[],[]) if iter else ([],[])
    else:
        raise IOError("a history filename or instance is required")
    ids = None
    try:  # read standard logfile (or monitor)
        if monitor:
            if iter:
                params, cost, ids = read_monitor(source, id=True)
                if not len(ids): ids = None
                ids = _process_ids(ids, len(cost))
            else:
                params, cost = read_monitor(source)
        elif data:
            params, cost = source.coords, source.values
            if iter: ids = _process_ids(source.ids, len(cost))
        elif arxiv: #NOTE: does not store ids
            params = list(list(k) for k in source.keys())
            cost = source.values()
            cost = cost if type(cost) is list else list(cost)
            if iter: ids = _process_ids(None, len(cost))
        elif solver: #NOTE: ids are single-valued, unless ensemble
            params, cost = source.solution_history, source.energy_history
            if iter: ids = _process_ids(source.id, len(cost))
        else:
            try:
                from mystic.solvers import LoadSolver
                return read_history(LoadSolver(source), iter)
            except: #KeyError
                ids, params, cost = logfile_reader(source, iter=True)
                if iter: ids = _process_ids(ids, len(cost))
                #FIXME: doesn't work for multi-id logfile; select id?
        params, cost = raw_to_support(params, cost)
    except:
        return read_raw_file(source+'.py', iter) #NOTE: assumes 'support' format
    return (ids, params, cost) if iter else (params, cost)


# logfile reader
#XXX: if have grad ([i,j,k]) not cost, 'cost' is list of lists [[i,j,k],...]
#XXX: if have grad and cost, 'cost' is mixed list [n,[i,j,k],m,[a,b,c],...]
#TODO: (mixed) create a function that splits/filters list/float in 'cost'
#TODO: (mixed) split/filter values in 'step','param' using indices from 'cost'
#TODO: behavior is identical for 'raw_to_support'... apply changes there also

def logfile_reader(filename, iter=False):
  """read a log file (e.g. written by a LoggingMonitor) in three-column format

  iter is a bool, where if True, `iter` is also returned
  filename is a log file path, with format:

  `__iter__  __energy__  __params__`

  `iter` is a tuple of (iteration, id), with `id` an int or None
  `energy` is a float or tuple[float], and is the output of the cost function
  `params` is a list[float], and is the input to the cost function

  If iter=True, returns tuple of (iter, params, energy), as defined above.
  If iter=False, returns tuple of (params, energy).
  """
  import numpy as np
  inf, nan = np.inf, np.nan
  f = open(filename,"r")
  file = f.read()
  f.close()
  contents = file.split("\n")
  # parse file contents to get (i,id), cost, and parameters
  step = []; cost = []; param = [];
  locals_ = locals().copy()
  for line in contents[:-1]:
    if line.startswith(("#","inf =","nan =")): pass
    else:
      values = line.split("   ")
      step.append(eval(values[0], {}, locals_))  #XXX: yields (i,id)
      cost.append(eval(values[1], {}, locals_))
      param.append(eval(values[2], {}, locals_))
  return (step, param, cost) if iter else (param, cost)

def read_trajectories(source, iter=False):
  """read trajectories from a monitor instance or three-column format log file

  iter is a bool, where if True, `iter` is also returned
  source is a monitor instance or a log file path, with format:

  `__iter__  __energy__  __params__`

  `iter` is a tuple of (iteration, id), with `id` an int or None
  `energy` is a float or tuple[float], and is the output of the cost function
  `params` is a list[float], and is the input to the cost function

  If iter=True, returns tuple of (iter, params, energy), as defined above.
  If iter=False, returns tuple of (params, energy).
  """
  if isinstance(source, str):
    return logfile_reader(source, iter)
  param, cost = source.x, source.y
  if not iter:
    return param, cost
  # otherwise, id may need some processing...
  return _process_ids(source.id, len(cost)), param, cost


def _reduce_ids(ids):
    """convert ids from list of tuples of (iterations, ids) to a list of ids
    """
    if not hasattr(ids, '__len__'):
        return ids
    if len(ids) and len(ids[0]) == 1:
        ids = [None]*len(ids)
    else:
        ids = [i[-1] for i in ids]
    return ids


def _process_ids(ids, n=None): #NOTE: n only needed when ids is single value
  """convert ids to list of tuples of (iterations, ids)

  ids is a list of ints, an int, or None
  n is target length (generally used when ids is a single-value)
  """
  if ids is None:
    return [(i,) for i in range(n)] if n else None
  from numbers import Integral
  if isinstance(ids, Integral):
    return [(i,ids) for i in range(n)] if n else ids
  if not len(ids):
    return [(i,) for i in range(n)] if n else []
  # ids is [(iter,id)] where id may be None
  if len(ids) and isinstance(ids[0], tuple):
    step = ids
  else:
    step = ids[:]
    for i,j in enumerate(ids):
      step[i] = (list(zip(*step[:i]))[1].count(j), j) if i else (i,j)
  if len(ids) == ids.count(None):
    step = [(i,) for (i,j) in step]
  else:
    step = list(step)
  return step[:n]


# read and write monitor (to and from raw data)

def read_monitor(mon, id=False):
  """read trajectories from a monitor instance

  mon is a monitor instance
  id is a bool, where if True, `id` is also returned

  Returns tuple of (mon.x, mon.y) or (mon.x, mon.y, mon.id)
  """
  steps = mon.x[:]
  energy = mon.y[:]
  if not id:
    return steps, energy
  id = mon.id[:]
  return steps, energy, id 

def write_monitor(steps, energy, id=None, k=None):
  """write trajectories to a monitor instance

  `steps` is a list[float], and is the input to the cost function
  `energy` is a float or tuple[float], and is the output of the cost function
  `id` is a list[int], and is the `id` of the monitored object
  `k` is float multiplier on `energy` (e.g. k=-1 inverts the cost function)

  Returns a mystic.monitors.Monitor instance
  """
  from mystic.monitors import Monitor
  mon = Monitor()
  mon.k = k
  mon._x.extend(steps)
  mon._y.extend(mon._k(energy, iter))
  if id is not None: mon._id.extend(id)
  return mon

# converters 

def converge_to_support(steps, energy):
  steps = zip(*steps)
  steps = [list(i) for i in steps]
  return steps, energy

def raw_to_converge(steps, energy):
  if len(steps) > 0:
    if not sequence(steps[0][0]):
      steps = [[step] for step in steps]  # needed when steps = [1,2,3,...]
    steps = [list(zip(*step)) for step in steps] # also can be used to revert 'steps'
  if len(energy) > 0:
    if hasattr(energy[0], 'tolist'):
      energy = [e.tolist() for e in energy]
  return steps, energy

def raw_to_support(steps, energy):
  return converge_to_support( *raw_to_converge(steps, energy) )

# monitor to file (support file, converge file, raw file)
## FIXME: 'converge' and 'raw' files are virtually unused and unsupported

def write_raw_file(mon,log_file='paramlog.py',**kwds):
  """write parameter and solution trajectory to a log file in 'raw' format

  mon is a monitor; log_file is a str log file path, with format:

  # header
  id = ...
  params = ...
  cost = ...

  `id` is a tuple of (iteration, id), with `id` an int or None
  `params` is a list[float], and is the input to the cost function
  `cost` is a float or tuple[float], and is the output of the cost function

  if header is provided, then the given string is written as the file header
  all other kwds are written as file entries
  """
  if isNull(mon): return  #XXX: throw error? warning? ???
  steps, energy, ids = read_monitor(mon, id=True) 
  if not len(ids): #XXX: is manipulating ids a good idea?
    ids = None
  elif ids.count(ids[0]) == len(ids): #XXX: generally, all None or all ints
    ids = ids[0]
  f = open(log_file,'w')
  if 'header' in kwds:
    f.write('# %s\n' % kwds['header'])
    del kwds['header']
  f.write("inf = float('inf')\n") # define special values
  f.write("nan = float('nan')\n") # define special values
  for variable,value in kwds.items():
    f.write('%s = %s\n' % (variable,value))# write remaining kwds as variables
  if ids is not None:
    f.write('id = %s\n' % ids)
 #f.write('# %s\n' % energy[-1])
  f.write('params = %s\n' % steps)
  f.write('cost = %s\n' % energy)
  f.close()
  return

def write_support_file(mon,log_file='paramlog.py',**kwds):
  """write parameter and solution trajectory to a log file in 'support' format

  mon is a monitor; log_file is a str log file path, with format:

  # header
  id = ...
  params = ...
  cost = ...

  `id` is a tuple of (iteration, id), with `id` an int or None
  `params` is a list[float], and is the input to the cost function
  `cost` is a float or tuple[float], and is the output of the cost function

  if header is provided, then the given string is written as the file header
  all other kwds are written as file entries

  NOTE: params are the transpose of how they are stored in monitor.x
  """
  if isNull(mon): return  #XXX: throw error? warning? ???
  monitor = write_monitor( *raw_to_support( *read_monitor(mon) ) )
  monitor._id = mon._id[:] #HACK: workaround loss of id above
  monitor.k = mon.k #HACK: workaround loss of k above (ensure is copy?)
  header = "written in 'support' format"
  if 'header' in kwds:
    header += "\n# " + str(kwds['header'])
    del kwds['header']
  write_raw_file(monitor,log_file,header=header,**kwds)
  return

def write_converge_file(mon,log_file='paramlog.py',**kwds):
  if isNull(mon): return  #XXX: throw error? warning? ???
  monitor = write_monitor( *raw_to_converge( *read_monitor(mon) ) )
  monitor._id = mon._id[:] #HACK: workaround loss of id above
  monitor.k = mon.k #HACK: workaround loss of k above (ensure is copy?)
  header = "written in 'converge' format"
  if 'header' in kwds:
    header += "\n# " + str(kwds['header'])
    del kwds['header']
  write_raw_file(monitor,log_file,header=header,**kwds)
  return

# file to data (support file, converge file, raw file)

def read_raw_file(file_in, iter=False):
    """read parameter and solution trajectory log file in 'raw' format

  file_in is a str log file path, with format:

  # header
  id = ...
  params = ...
  cost = ...

  `id` is a tuple of (iteration, id), with `id` an int or None
  `params` is a list[float], and is the input to the cost function
  `cost` is a float or tuple[float], and is the output of the cost function

  if iter is True, return (id,params,cost) otherwise return (params,cost)

  NOTE: params are stored how they are stored in monitor.x
    """
    if iter:
        ids, params, cost = read_import(file_in,'id','params','cost')
        return _process_ids(ids, len(cost)), params, cost
    else:
        return read_import(file_in,'params','cost')

#TODO: check impact of having gradient ([i,j,k]) and/not cost
def read_import(file, *targets):
  "import the targets; targets are name strings"
  import re, os, sys
  _dir, file = os.path.split(file)
  file = re.sub(r'\.py*.$', '', file) #XXX: strip .py* extension
  curdir = os.path.abspath(os.curdir)
  sys.path.append('.')
  results = []
  globals = {}
  try:
    if _dir: os.chdir(_dir)
    if len(targets):
      for target in targets:
        code = "from {0} import {1} as result".format(file, target)
        code = compile(code, '<string>', 'exec')
        try:
            exec(code, globals)
        except ModuleNotFoundError:
            raise RuntimeError('Module: {0} not found'.format(file))
        except ImportError:
            globals['result'] = None #XXX: or throw error?
        results.append(globals['result'])
    else:
        code = "import {0} as result".format(file)
        code = compile(code, '<string>', 'exec')
        try:
            exec(code, globals)
        except ModuleNotFoundError:
            raise RuntimeError('Module: {0} not found'.format(file))
        except ImportError:
            globals['result'] = None #XXX: or throw error?
        results.append(globals['result'])
  except FileNotFoundError:
    raise RuntimeError('File: {0} not found'.format(file))
  finally:
    if _dir: os.chdir(curdir)
    sys.path.pop()
  if not len(results): return None
  return results[-1] if (len(results) == 1) else results

def read_converge_file(file_in, iter=False):
  data = read_raw_file(file_in, iter)
  return (data[0], raw_to_converge(*data[1:])) if iter else raw_to_converge(*data)
 #return ids, raw_to_converge(params,cost)
 #params = [zip(*param) for param in params] # alternate to revert 'params'

def read_support_file(file_in, iter=False):
  """read parameter and solution trajectory log file in 'support' format

  file_in is a str log file path, with format:

  # header
  id = ...
  params = ...
  cost = ...

  `id` is a tuple of (iteration, id), with `id` an int or None
  `params` is a list[float], and is the input to the cost function
  `cost` is a float or tuple[float], and is the output of the cost function

  if iter is True, return (id,params,cost) otherwise return (params,cost)

  NOTE: params are the transpose of how they are stored in monitor.x
  """
  data = read_raw_file(file_in, iter)
  return (data[0], raw_to_support(*data[1:])) if iter else raw_to_support(*data)
 #return ids, raw_to_support(params,cost)

# file converters

def raw_to_converge_converter(file_in,file_out):
  steps, energy = read_raw_file(file_in)
  write_raw_file(write_monitor( *raw_to_converge(steps,energy) ),
                 file_out,header="written in 'converge' format")
  return

def raw_to_support_converter(file_in,file_out):
  steps, energy = read_raw_file(file_in)
  write_raw_file(write_monitor( *raw_to_support(steps,energy) ),
                 file_out,header="written in 'support' format")
  return

def converge_to_support_converter(file_in,file_out):
  steps, energy = read_converge_file(file_in)
  write_raw_file(write_monitor( *converge_to_support(steps,energy) ),
                 file_out,header="written in 'support' format")
  return

# old #

def read_old_support_file(file_in):
  steps, energy = read_raw_file(file_in)
  steps = [[(i,) for i in steps[j]] for j in range(len(steps))]
  energy = [(i,) for i in energy]
  return write_monitor(steps, energy)

def old_to_new_support_converter(file_in,file_out):
  mon = read_old_support_file(file_in)
  write_raw_file(mon,file_out)
  return

def __orig_write_support_file(mon,log_file='paramlog.py'):
  if isNull(mon): return  #XXX: throw error? warning? ???
  steps, energy = read_monitor(mon)
  log = []
  if len(steps) > 0:
    for p in range(len(steps[0])):
      q = []
      for s in range(len(steps)):
        q.append(steps[s][p])
      log.append(q)  
  monitor = write_monitor(log, energy)
  write_raw_file(monitor,log_file)
  return

def __orig_write_converge_file(mon,log_file='paramlog.py'):
  write_raw_file(mon,log_file)
  return


if __name__ == '__main__':
  pass

# EOF



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

def objective(x):
    from numpy import abs, sum, cos, prod as product, sqrt
    sum_jx = 0.0
    for j in range(len(x)): sum_jx = sum_jx + (j+1) * x[j]**2
    return -abs((sum(cos(x)**4) - 2.*product(cos(x)**2))/sqrt(sum_jx))

def bounds(len=3):
    return [(0.0,10.0)]*len

# with penalty='penalty' applied, solution is:
xs = [ 3.04295933, 1.48286963, 0.16618875]
ys = -0.51578987

"""
for len(x) = 10,
xs ~ [3.12388714, 3.06913834, 3.01426760, 2.95755412, 1.46603517,
      0.36802963, 0.36346912, 0.35912472, 0.35493945, 0.35095372]
ys ~ -0.74732020
"""
bounds = bounds(len(xs))

from mystic.constraints import and_
from mystic.symbolic import (generate_constraint, generate_solvers,
                             generate_penalty, generate_conditions,
                             simplify, symbolic_bounds)

equations = """
-prod([x0, x1, x2]) + 0.75 <= 0.0
sum([x0, x1, x2]) - 7.5*3 <= 0.0
"""
equations += symbolic_bounds(*zip(*bounds))
cf = generate_constraint(generate_solvers(simplify(equations)), join=and_)
pf = generate_penalty(generate_conditions(equations))



if __name__ == '__main__':
    from mystic.solvers import diffev2
    from mystic.math import almostEqual

    result = diffev2(objective, x0=bounds, bounds=bounds, constraints=cf, penalty=pf, npop=80, disp=False, full_output=True, gtol=100)

    assert almostEqual(result[0], xs, rel=1e-1)
    assert almostEqual(result[1], ys, rel=1e-1)



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
calculate lower and upper bound on expected Error on |truth - surrogate|

Test function is y = F(x), where:
  y0 = x0 + x1 * | x2 * x3**2 - (x4 / x1)**2 |**.5
  y1 = x0 - x1 * | x2 * x3**2 + (x4 / x1)**2 |**.5
  y2 = x0 - | x1 * x2 * x3 - x4 |

toy = lambda x: F(x)[0]
truth = lambda x: toy(x + .01) - .01
surrogate is trained on data sampled from truth
error = lambda x: (truth(x) - surrogate(x))**2

Calculate lower and upper bound on E|error(x)|, where:
  x in [(0,1), (1,10), (0,10), (0,10), (0,10)]
  wx in [(0,1), (1,1), (1,1), (1,1), (1,1)]
  npts = [2, 1, 1, 1, 1] (i.e. two Dirac masses on x[0], one elsewhere)
  sum(wx[i]_j) for j in [0,npts], for each i
  mean(x[0]) = 5e-1 +/- 1e-3
  var(x[0]) = 5e-3 +/- 1e-4

Solves for two scenarios of x that produce lower bound on E|error(x)|,
given the bounds, normalization, and moment constraints. Repeats
calculation for upper bound on E|error(x)|.

Creates 'log' of optimizations and 'truth' database of stored evaluations.
'''
from ouq_models import *


if __name__ == '__main__':

    #from toys import cost5x3 as toy; nx = 5; ny = 3
    #from toys import function5x3 as toy; nx = 5; ny = 3
    #from toys import cost5x1 as toy; nx = 5; ny = 1
    #from toys import function5x1 as toy; nx = 5; ny = 1
    #from toys import cost5 as toy; nx = 5; ny = None
    from toys import function5 as toy; nx = 5; ny = None

    # update optimization parameters
    from misc import param, npts, wlb, wub, is_cons, scons
    from ouq import ExpectedValue
    from mystic.bounds import MeasureBounds
    from mystic.monitors import VerboseLoggingMonitor, Monitor, VerboseMonitor
    from mystic.termination import VTRChangeOverGeneration as VTRCOG
    from mystic.termination import Or, VTR, ChangeOverGeneration as COG
    param['opts']['termination'] = COG(1e-10, 100) #NOTE: short stop?
    param['npop'] = 80 #NOTE: increase if poor convergence
    param['stepmon'] = VerboseLoggingMonitor(1, 20, filename='log.txt', label='lower')

    # build bounds
    bnd = MeasureBounds((0,1,0,0,0)[:nx],(1,10,10,10,10)[:nx], n=npts[:nx], wlb=wlb[:nx], wub=wub[:nx])

    # build a model representing 'truth', and generate some data
    #print("building truth F'(x|a')...")
    true = dict(mu=.01, sigma=0., zmu=-.01, zsigma=0.)
    truth = NoisyModel('truth', model=toy, nx=nx, ny=ny, **true)
    #print('sampling truth...')
    data = truth.sample([(0,1),(1,10)]+[(0,10)]*(nx-2), pts='16')
    Ns = 25 #XXX: number of samples, when model has randomness

    try: # parallel maps
        from pathos.maps import Map
        from pathos.pools import ThreadPool, _ThreadPool
        pmap = Map(ThreadPool) if Ns else Map() # for sampling
        if ny: param['axmap'] = Map(_ThreadPool, join=True) # for multi-axis
    except ImportError:
        pmap = None

    # build a surrogate model by training on the data
    args = dict(alpha=1e-10, optimizer=None, n_restarts_optimizer=0)
    #args['optimizer'] = 'fmin_l_bfgs_b'
    kargs = dict(length_scale=1.0, length_scale_bounds=(1e-05, 100000.0))
    wargs = dict(noise_level=1.0, noise_level_bounds=(1e-05, 100000.0))
    from sklearn.gaussian_process import GaussianProcessRegressor as GPRegressor
    from sklearn.gaussian_process.kernels import RBF, WhiteKernel
    from sklearn.preprocessing import StandardScaler
    from ml import Estimator, MLData, improve_score
    args['kernel'] = RBF(**kargs)#+ WhiteKernel(**wargs) #XXX: use if noisy
    kwds = dict(estimator=GPRegressor(**args), transform=StandardScaler())
    # iteratively improve estimator
    gpr = Estimator(**kwds) #FIXME: traintest so train != test ?
    best = improve_score(gpr, MLData(data.coords, data.coords, data.values, data.values), tries=10, verbose=True)
    mlkw = dict(estimator=best.estimator, transform=best.transform)

    #print('building estimator G(x) from truth data...')
    surrogate = LearnedModel('surrogate', nx=nx, ny=ny, data=truth, **mlkw)
    #print('building UQ model of model error...')
    error = ErrorModel('error', model=truth, surrogate=surrogate)

    rnd = Ns if error.rnd else None
    if not rnd: pmap = None
    #print('building UQ objective of expected model error...')
    b = ExpectedValue(error, bnd, constraint=scons, cvalid=is_cons, samples=rnd, map=pmap)
    #print('solving for lower bound on expected model error...')
    b.lower_bound(axis=None, id=0, **param)
    print("lower bound per axis:")
    for axis,solver in b._lower.items():
        print("%s: %s @ %s" % (axis, solver.bestEnergy, solver.bestSolution))

    #print('solving for upper bound on expected model error...')
    param['opts']['termination'] = COG(1e-10, 200) #NOTE: short stop?
    param['npop'] = 160 #NOTE: increase if poor convergence
    param['stepmon'] = VerboseLoggingMonitor(1, 20, filename='log.txt', label='upper')
    b.upper_bound(axis=None, id=1, **param)
    print("upper bound per axis:")
    for axis,solver in b._upper.items():
        print("%s: %s @ %s" % (axis, -solver.bestEnergy, solver.bestSolution))



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
a global searcher
"""
#FIXME: refactor to use mystic.samplers (and code in _workflow)

class Searcher(object):
   #searcher has:
   #    sprayer - an ensemble algorithm
   #    archive - a sampled point archive(s)
   #    cache - a trajectory cache(s)
   #    retry - max consectutive retries w/o an archive 'miss'
   #    repeat - number of times to repeat the search
   #    tol - minima comparator rounding precision
   #    _allSolvers - collection of sprayers (for all solver trajectories)
   #
   #searcher (or archive) has:
   #    memtol - memoization rounding precision
   #
   #searcher (or sprayer) has:
   #    seeker - an optimiztion algorithm
   #    map - an enselble map
   #    traj - trajectory configuration (bool)
   #    disp - verbosity configuration (bool)
   #    npts - number of solvers
   #
   #searcher can:
   #    _memoize - apply caching archive to sprayer
   #    _search - perform a single search iteration (the search algorithm ?)
   #    Search - keep performing _search until retry condition is met
   #    UseTrajectories - save all sprayers, thus save all trajectories
   #    Reset - clear caching archive and saved sprayers (thus trajectories)
   #    Trajectories - fetch (step, param, cost) for all solver trajectories
   #    Samples - all sampled points as a n-D array (output is array[-1])
   #    Values - fetch values (model(x)) for sampled points on surface (model)
   #    Coordinates - fetch inputs (x) for sampled points on surface (model)
   #    Minima - fetch {x: model(x)} for all x that minimize model
   #    _summarize - print diagnostics on the value/quantity of solved values
   #
   # searcher (or sprayer) can:
   #    _print - print bestSolution and bestEnergy for each sprayer
   #    _configure (model, bounds, stop, monitor) - configure sprayer
   #    _solve - spray multiple seekers

    def __init__(self, npts=4, retry=1, tol=8, memtol=1, memsize=None,
                       map=None, archive=None, cache=None, sprayer=None,
                       seeker=None, traj=False, disp=False, repeat=0):
        """searcher, which searches for all minima of a response surface

    Input:
      npts - number of solvers in the ensemble
      retry - max consectutive retries w/o a cache 'miss'
      tol - rounding precision for the minima comparator
      memtol - rounding precision for memoization
      memsize - maximum size of cache to hold in memory
      map - map used for spawning solvers in the ensemble
      archive - the sampled point archive(s)
      cache - the trajectory cache(s)
      sprayer - the mystic.ensemble instance
      seeker - the mystic.solvers instance
      traj - if True, save the parameter trajectories
      disp - if True, be verbose
      repeat - number of times to repeat the search
        """
        #XXX: better not to use klepto as default? just dict and false cache?
        from klepto.archives import dict_archive as _archive
        from mystic.solvers import BuckshotSolver #XXX: or SparsitySolver?
        from mystic.solvers import PowellDirectionalSolver
        from mystic.pools import SerialPool as Pool
        from builtins import map as _map
        self.archive = _archive(cached=False) if archive is None else archive
        self.cache = _archive(cached=False) if cache is None else cache
        self.sprayer = BuckshotSolver if sprayer is None else sprayer
        self.seeker = PowellDirectionalSolver if seeker is None else seeker
        self.map = Pool().map if map in (None, _map) else map 
        self.traj = traj
        self.disp = disp

        self.npts = npts # number of solvers
        self.retry = retry   # max consectutive retries w/o a cache 'miss'
        self.repeat = repeat # number of times to repeat the search
        self.tol = tol       # rounding precision
        self.memtol = memtol # memoization rounding precision
        self.memsize = memsize # max in-memory cache size
        self._allSolvers = []
        self._inv = False    # self-awareness: am I inverted (maximizing)?
        return

    def _print(self, solver, tol=8):
        """print bestSolution and bestEnergy for each sprayer"""
        l = -1 if self._inv else 1
        for _solver in solver._allSolvers:
            bestSol = tuple(round(s, tol) for s in _solver.bestSolution)
            bestRes = round(_solver.bestEnergy, tol)
            print("%s %s" % (bestSol, l*bestRes))

    def _memoize(self, solver, tol=1, all=False, size=None):
        """apply caching to ensemble solver instance"""
        archive = self.archive if all else self.cache

        from klepto import lru_cache as _cache #XXX: or lru_? or ?
        from klepto.keymaps import keymap

        km = keymap()
        ca = _cache(maxsize=size, ignore=('**','out'),
                    tol=tol, cache=archive, keymap=km)

        @ca
        def memo(*args, **kwds):
            return kwds['out']

        l = -1 if self._inv else 1
        if all: #FIXME: applied *after* _solve; should be *during* _solve
            cache = memo.__cache__()
            for _solver in solver._allSolvers:
                if _solver._evalmon:
                    param = _solver._evalmon._x
                    cost = _solver._evalmon._y
                    cache.update((tuple(x),l*y) for (x,y) in zip(param,cost))
        else:
            for _solver in solver._allSolvers:
                bestSol = tuple(_solver.bestSolution)
                bestRes = float(_solver.bestEnergy)
                memo(*bestSol, out=l*bestRes)
        return memo

    #FIXME: instead, take a configured solver?
    def _configure(self, model, bounds, stop=None, **kwds):
        """configure ensemble solver from objective and other solver options"""
        monitor = kwds.get('monitor', None)
        evalmon = kwds.get('evalmon', None)
        penalty = kwds.get('penalty', None)
        constraints = kwds.get('constraints', None)
        tight = kwds.get('tightrange', None)
        clip = kwds.get('cliprange', None)

        from mystic.monitors import Monitor
        # configure monitor
        monitor = Monitor() if monitor is None else monitor

        # get dimensions
        ndim = len(bounds)
        _min, _max = zip(*bounds)

        # configure ensemble solver
        self.solver = self.sprayer(ndim, self.npts)
        self.solver.SetNestedSolver(self.seeker)
        self.solver.SetMapper(self.map)
        self.solver.SetObjective(model)
        self.solver.SetConstraints(constraints)
        self.solver.SetPenalty(penalty)
        if stop: self.solver.SetTermination(stop)
        if evalmon is not None: self.solver.SetEvaluationMonitor(evalmon)
        if monitor is not None: self.solver.SetGenerationMonitor(monitor) #NOTE: may be xy,-z
        self.solver.SetStrictRanges(min=_min, max=_max, tight=tight, clip=clip)
        return

    def _solve(self, id=None, disp=None):
        """run the solver (i.e. search for the minima)"""
        from copy import deepcopy as _copy
        solver = _copy(self.solver)
        # configure solver
        solver.id = id
        model = solver._cost[1] #FIXME: HACK b/c Solve(model) is required
        # solve
        disp = self.disp if disp is None else disp
#       import time
#       start = time.time()
        solver.Solve(model, disp=disp)
#       print("TOOK: %s" % (time.time() - start))
        return solver

    def _search(self, sid): #FIXME: load and leverage archived evaluations
        """run the solver, store the trajectory, and cache to the archive"""
        solver = self._solve(sid, self.disp)
        if self.traj: self._allSolvers.append(solver)
        sid += len(solver._allSolvers)
#       self._print(solver, tol=self.tol)
        size = self.memsize
        # write to evaluation cache
        memo = self._memoize(solver, tol=None, all=True, size=size).info()
        # write to trajectory archive (extrema only)
        info = self._memoize(solver, tol=self.memtol).info()
        if self.disp: print(info)
        size = info.size
        return sid, size

    def UseTrajectories(self, traj=True):
        """save all sprayers, thus save all trajectories
        """
        self.traj = bool(traj)
        return

    def Verbose(self, disp=True):
        """be verbose
        """
        self.disp = bool(disp)
        return

    #FIXME: instead, take a configured solver?
    def Search(self, model, bounds, stop=None, traj=None, disp=None, **kwds):
        """use an ensemble of optimizers to search for all minima

    Input:
      model - function z=f(x) to be used as the objective of the Searcher
      bounds - mystic.bounds instance on the search region
      stop - termination condition
      traj - klepto.archive to store sampled points
      disp - if True, be verbose
      monitor - mystic.monitor instance to store parameter trajectories
      evalmon - mystic.monitor instance to store parameter evaluations
      penalty - mystic.penalty instance of the form y' = k*p(x)
      constraints - mystic.constraints instance of the form x' = c(x)
      tightrange - if True, apply bounds concurrent with other constraints
      cliprange - if True, bounding constraints will clip exterior values
        """
        self.traj = self.traj if traj is None else traj
        self.disp = self.disp if disp is None else disp
        self._configure(model, bounds, stop, **kwds)
        sid = 0  # keep track of which solver is which across multiple runs
        run = -1
        while run < self.repeat: # stop after repeat 'runs'
            #print('run: {}'.format(run))
            count = 0 if self.retry else -1 
            while self.retry > count: # stop after retry consecutive no new hits
                #print('count: {}'.format(count))
                _size = -1
                size = osize = len(self.cache) #XXX: 'size' or 'len(vals)'?
                while size > _size: # stop if no new hits
                    #print('size: {}, _: {}'.format(size,_size))
                    _size = size
                    sid, size = self._search(sid) # uses self.traj and self.disp
                if size == osize: count = count + 1
                else: count = 0
            run = run + 1

        #NOTE: traj & disp are sticky
        return

    def Reset(self, cache=None, inv=None):
        """clear the trajectory cache of sampled points

    Input:
      cache - the trajectory cache(s)
      inv - if True, reset the cache for the inverse of the objective
        """
        if cache is None: self.cache.clear() #XXX: clear the archive?
        self.cache = self.cache if cache is None else cache
        [self._allSolvers.pop() for i in range(len(self._allSolvers))]
        if inv is not None: self._inv = inv

    def Values(self, unique=False, all=False):
        """return the sequence of stored response surface outputs

    Input:
      unique: if True, only return unique values
      all: if True, return all sampled values (not just trajectory values)

    Returns:
      a list of stored response surface outputs
        """
        archive = self.archive if all else self.cache
        vals = archive.values()
        new = set()
        return [v for v in vals if v not in new and not new.add(v)] if unique else list(vals)

    def Coordinates(self, unique=False, all=False):
        """return the sequence of stored model parameter input values

    Input:
      unique: if True, only return unique values
      all: if True, return all sampled inputs (not just trajectory inputs)

    Returns:
      a list of parameter trajectories
        """
        archive = self.archive if all else self.cache
        keys = archive.keys()
        new = set()
        return [k for k in keys if k not in new and not new.add(k)] if unique else list(keys)

    def Minima(self, tol=None): #XXX: unique?
        """return a dict of (coordinates,values) of all discovered minima

    Input:
      tol: tolerance within which to consider a point a minima

    Returns:
      a dict of (coordinates,values) of all discovered minima
        """
        if tol is None: tol=self.tol
        data = self.cache
        _min = max if self._inv else min
        _min = _min(data.values())
        return dict((k,v) for (k,v) in data.items() if round(v, tol) == round(_min, tol))

    def _summarize(self):
        """provide a summary of the search results"""
        from builtins import min as _min
        #NOTE: len(size) = # of dirs; len(vals) = # unique dirs
        keys = self.Coordinates()
        vals = self.Values()
        mins = self.Minima()
        name = 'max' if self._inv else 'min'
        min = max if self._inv else _min
        # print the minimum and number of times the minimum was found
        print("%s: %s (count=%s)" % (name, min(mins.values()), len(mins)))
        # print number of minima found, number of unique minima, archive size
        print("pts: %s (values=%s, size=%s)" % (len(set(keys)), len(set(vals)), len(keys)))
        #_i = max(dict(i).values())
        return

    def Trajectories(self, all=False):
        """return tuple (iteration, coordinates, cost) of all trajectories
        """
        mon = '_evalmon' if all else '_stepmon'
        from mystic.munge import read_trajectories
        if not self.traj:
            try: #NOTE: FRAGILE (if absolute path is not used)
                filename = getattr(self.solver, mon)._filename
                step, param, cost = read_trajectories(filename, iter=True)
            except AttributeError:
                msg = "a LoggingMonitor or UseTrajectories is required"
                raise RuntimeError(msg)
        else:
            step = []; cost = []; param = [];
            for sprayer in self._allSolvers:  #XXX: slow? better thread.map?
                for seeker in sprayer._allSolvers:
                    values = read_trajectories(getattr(seeker,mon), iter=True)
                    step.extend(values[0])
                    param.extend(values[1])
                    cost.extend(values[2])
        #XXX: (not from archive, so) if self._inv: use -cost
        return step, param, cost

    def Samples(self, all=False):
        """return array of (coordinates, cost) for all trajectories
        """
        import numpy as np
        xy,xy,z = self.Trajectories(all=all)
        xy = np.vstack((np.array(xy).T,z))
        if self._inv: xy[-1,:] = -xy[-1]
        return xy #NOTE: actually xyz

    pass



# EOF
